from __future__ import annotations

from abc import ABC
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import date, datetime
from functools import cached_property, singledispatchmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    List,
    Optional,
    Self,
    Sequence,
    Set,
    Tuple,
    Union,
)

from pydantic import (
    ConfigDict,
)

from trilogy.constants import DEFAULT_NAMESPACE, MagicConstants
from trilogy.core.constants import ALL_ROWS_CONCEPT
from trilogy.core.enums import (
    BooleanOperator,
    ComparisonOperator,
    DatePart,
    Derivation,
    FunctionClass,
    FunctionType,
    Granularity,
    Modifier,
    Ordering,
    Purpose,
    WindowType,
)
from trilogy.core.models.author import (
    AggregateWrapper,
    AlignClause,
    AlignItem,
    ArgBinding,
    CaseElse,
    CaseWhen,
    Comparison,
    Concept,
    ConceptRef,
    Conditional,
    FilterItem,
    FuncArgs,
    Function,
    FunctionCallWrapper,
    Grain,
    HavingClause,
    Metadata,
    MultiSelectLineage,
    OrderBy,
    OrderItem,
    Parenthetical,
    RowsetItem,
    RowsetLineage,
    SelectLineage,
    SubselectComparison,
    UndefinedConcept,
    WhereClause,
    WindowItem,
)
from trilogy.core.models.core import (
    Addressable,
    ArrayType,
    DataType,
    DataTyped,
    ListWrapper,
    MapType,
    MapWrapper,
    NumericType,
    StructType,
    TraitDataType,
    TupleWrapper,
    arg_to_datatype,
)
from trilogy.core.models.datasource import (
    Address,
    ColumnAssignment,
    Datasource,
    DatasourceMetadata,
    RawColumnExpr,
)
from trilogy.core.models.environment import Environment

# TODO: refactor to avoid these
if TYPE_CHECKING:
    from trilogy.core.models.build_environment import BuildEnvironment
    from trilogy.core.models.execute import CTE, UnionCTE

LOGGER_PREFIX = "[MODELS_BUILD]"


class BuildConceptArgs(ABC):
    @property
    def concept_arguments(self) -> Sequence["BuildConcept"]:
        raise NotImplementedError

    @property
    def existence_arguments(self) -> Sequence[tuple["BuildConcept", ...]]:
        return []

    @property
    def row_arguments(self) -> Sequence["BuildConcept"]:
        return self.concept_arguments


def concept_is_relevant(
    concept: BuildConcept,
    others: list[BuildConcept],
) -> bool:

    if concept.is_aggregate and not (
        isinstance(concept.lineage, BuildAggregateWrapper) and concept.lineage.by
    ):

        return False
    if concept.purpose in (Purpose.PROPERTY, Purpose.METRIC) and concept.keys:
        if all([c in others for c in concept.keys]):
            return False
    if (
        concept.purpose == Purpose.KEY
        and concept.keys
        and all([c in others and c != concept.address for c in concept.keys])
    ):
        return False
    if concept.purpose in (Purpose.METRIC,):
        if all([c in others for c in concept.grain.components]):
            return False
    if concept.derivation in (Derivation.BASIC,):

        return any(concept_is_relevant(c, others) for c in concept.concept_arguments)
    if concept.granularity == Granularity.SINGLE_ROW:
        return False
    return True


def concepts_to_build_grain_concepts(
    concepts: Iterable[BuildConcept | str], environment: "BuildEnvironment" | None
) -> set[str]:
    pconcepts: list[BuildConcept] = []
    for c in concepts:
        if isinstance(c, BuildConcept):
            pconcepts.append(c)
        elif environment:
            pconcepts.append(environment.concepts[c])

        else:
            raise ValueError(
                f"Unable to resolve input {c} without environment provided to concepts_to_grain call"
            )

    final: set[str] = set()
    for sub in pconcepts:
        if not concept_is_relevant(sub, pconcepts):
            continue
        final.add(sub.address)

    return final


@dataclass
class LooseBuildConceptList:
    concepts: Sequence[BuildConcept]

    @cached_property
    def addresses(self) -> set[str]:
        return {s.address for s in self.concepts}

    @cached_property
    def sorted_addresses(self) -> List[str]:
        return sorted(list(self.addresses))

    def __str__(self) -> str:
        return f"lcl{str(self.sorted_addresses)}"

    def __iter__(self):
        return iter(self.concepts)

    def __eq__(self, other):
        if not isinstance(other, LooseBuildConceptList):
            return False
        return self.addresses == other.addresses

    def issubset(self, other):
        if not isinstance(other, LooseBuildConceptList):
            return False
        return self.addresses.issubset(other.addresses)

    def __contains__(self, other):
        if isinstance(other, str):
            return other in self.addresses
        if not isinstance(other, BuildConcept):
            return False
        return other.address in self.addresses

    def difference(self, other):
        if not isinstance(other, LooseBuildConceptList):
            return False
        return self.addresses.difference(other.addresses)

    def isdisjoint(self, other):
        if not isinstance(other, LooseBuildConceptList):
            return False
        return self.addresses.isdisjoint(other.addresses)


class ConstantInlineable(ABC):

    def inline_constant(self, concept: BuildConcept):
        raise NotImplementedError


def get_concept_row_arguments(expr) -> List["BuildConcept"]:
    output = []
    if isinstance(expr, BuildConcept):
        output += [expr]

    elif isinstance(expr, BuildConceptArgs):
        output += expr.row_arguments
    return output


def get_concept_arguments(expr) -> List["BuildConcept"]:
    output = []
    if isinstance(expr, BuildConcept):
        output += [expr]

    elif isinstance(
        expr,
        BuildConceptArgs,
    ):
        output += expr.concept_arguments
    return output


@dataclass
class BuildParamaterizedConceptReference(DataTyped):
    concept: BuildConcept

    def __str__(self):
        return f":{self.concept.address}"

    @property
    def safe_address(self) -> str:
        return self.concept.safe_address

    @property
    def output_datatype(self) -> DataType:
        return self.concept.output_datatype


@dataclass
class BuildGrain:
    components: set[str] = field(default_factory=set)
    where_clause: Optional[BuildWhereClause] = None
    _str: str | None = None
    _str_no_condition: str | None = None

    def without_condition(self):
        if not self.where_clause:
            return self
        return BuildGrain(components=self.components)

    @classmethod
    def from_concepts(
        cls,
        concepts: Iterable[BuildConcept | str],
        environment: BuildEnvironment | None = None,
        where_clause: BuildWhereClause | None = None,
    ) -> "BuildGrain":

        return BuildGrain(
            components=concepts_to_build_grain_concepts(
                concepts, environment=environment
            ),
            where_clause=where_clause,
        )

    def __add__(self, other: "BuildGrain") -> "BuildGrain":
        if not other:
            return self
        where = self.where_clause
        if other.where_clause:
            if not self.where_clause:
                where = other.where_clause
            elif not other.where_clause == self.where_clause:
                where = BuildWhereClause(
                    conditional=BuildConditional(
                        left=self.where_clause.conditional,
                        right=other.where_clause.conditional,
                        operator=BooleanOperator.AND,
                    )
                )
                # raise NotImplementedError(
                #     f"Cannot merge grains with where clauses, self {self.where_clause} other {other.where_clause}"
                # )
        return BuildGrain(
            components=self.components.union(other.components), where_clause=where
        )

    def __sub__(self, other: "BuildGrain") -> "BuildGrain":
        return BuildGrain(
            components=self.components.difference(other.components),
            where_clause=self.where_clause,
        )

    @property
    def abstract(self):
        return not self.components or all(
            [c.endswith(ALL_ROWS_CONCEPT) for c in self.components]
        )

    def __eq__(self, other: object):
        if isinstance(other, list):
            if all([isinstance(c, BuildConcept) for c in other]):
                return self.components == set([c.address for c in other])
            return False
        if not isinstance(other, BuildGrain):
            return False
        if self.components == other.components:
            return True
        return False

    def issubset(self, other: "BuildGrain"):
        return self.components.issubset(other.components)

    def union(self, other: "BuildGrain"):
        addresses = self.components.union(other.components)
        return BuildGrain(components=addresses, where_clause=self.where_clause)

    def isdisjoint(self, other: "BuildGrain"):
        return self.components.isdisjoint(other.components)

    def intersection(self, other: "BuildGrain") -> "BuildGrain":
        intersection = self.components.intersection(other.components)
        return BuildGrain(components=intersection)

    def _calculate_string(self):
        if self.abstract:
            base = "Grain<Abstract>"
        else:
            base = "Grain<" + ",".join(sorted(self.components)) + ">"
        if self.where_clause:
            base += f"|{str(self.where_clause)}"
        return base

    def _calculate_string_no_condition(self):
        if self.abstract:
            base = "Grain<Abstract>"
        else:
            base = "Grain<" + ",".join(sorted(self.components)) + ">"
        return base

    @property
    def str_no_condition(self):
        if self._str_no_condition:
            return self._str_no_condition
        self._str_no_condition = self._calculate_string_no_condition()
        return self._str_no_condition

    def __str__(self):
        if self._str:
            return self._str
        self._str = self._calculate_string()
        return self._str

    def __radd__(self, other) -> "BuildGrain":
        if other == 0:
            return self
        else:
            return self.__add__(other)


@dataclass
class BuildParenthetical(DataTyped, ConstantInlineable, BuildConceptArgs):
    content: "BuildExpr"

    def __add__(self, other) -> Union["BuildParenthetical", "BuildConditional"]:
        if other is None:
            return self
        elif isinstance(other, (BuildComparison, BuildConditional, BuildParenthetical)):
            return BuildConditional(
                left=self, right=other, operator=BooleanOperator.AND
            )
        raise ValueError(f"Cannot add {self.__class__} and {type(other)}")

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"({str(self.content)})"

    def inline_constant(self, BuildConcept: BuildConcept):
        return BuildParenthetical(
            content=(
                self.content.inline_constant(BuildConcept)
                if isinstance(self.content, ConstantInlineable)
                else self.content
            )
        )

    @property
    def concept_arguments(self) -> List[BuildConcept]:
        base: List[BuildConcept] = []
        x = self.content
        if isinstance(x, BuildConcept):
            base += [x]
        elif isinstance(x, BuildConceptArgs):
            base += x.concept_arguments
        return base

    @property
    def row_arguments(self) -> Sequence[BuildConcept]:
        if isinstance(self.content, BuildConceptArgs):
            return self.content.row_arguments
        return self.concept_arguments

    @property
    def existence_arguments(self) -> Sequence[tuple["BuildConcept", ...]]:
        if isinstance(self.content, BuildConceptArgs):
            return self.content.existence_arguments
        return []

    @property
    def output_datatype(self):
        return arg_to_datatype(self.content)


@dataclass
class BuildConditional(BuildConceptArgs, ConstantInlineable):
    left: Union[
        int,
        str,
        float,
        list,
        bool,
        MagicConstants,
        BuildConcept,
        BuildComparison,
        BuildConditional,
        BuildParenthetical,
        BuildSubselectComparison,
        BuildFunction,
        BuildFilterItem,
    ]
    right: Union[
        int,
        str,
        float,
        list,
        bool,
        MagicConstants,
        BuildConcept,
        BuildComparison,
        BuildConditional,
        BuildParenthetical,
        BuildSubselectComparison,
        BuildFunction,
        BuildFilterItem,
    ]
    operator: BooleanOperator

    def __add__(self, other) -> "BuildConditional":
        if other is None:
            return self
        elif str(other) == str(self):
            return self
        elif isinstance(other, (BuildComparison, BuildConditional, BuildParenthetical)):
            return BuildConditional(
                left=self, right=other, operator=BooleanOperator.AND
            )
        raise ValueError(f"Cannot add {self.__class__} and {type(other)}")

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"{str(self.left)} {self.operator.value} {str(self.right)}"

    def __eq__(self, other):
        if not isinstance(other, BuildConditional):
            return False
        return (
            self.left == other.left
            and self.right == other.right
            and self.operator == other.operator
        )

    def inline_constant(self, constant: BuildConcept) -> "BuildConditional":
        assert isinstance(constant.lineage, BuildFunction)
        new_val = constant.lineage.arguments[0]
        if isinstance(self.left, ConstantInlineable):
            new_left = self.left.inline_constant(constant)
        elif (
            isinstance(self.left, BuildConcept)
            and self.left.address == constant.address
        ):
            new_left = new_val
        else:
            new_left = self.left

        if isinstance(self.right, ConstantInlineable):
            new_right = self.right.inline_constant(constant)
        elif (
            isinstance(self.right, BuildConcept)
            and self.right.address == constant.address
        ):
            new_right = new_val
        else:
            new_right = self.right

        if self.right == constant:
            new_right = new_val

        return BuildConditional(
            left=new_left,
            right=new_right,
            operator=self.operator,
        )

    @property
    def concept_arguments(self) -> List[BuildConcept]:
        """Return BuildConcepts directly referenced in where clause"""
        output = []
        output += get_concept_arguments(self.left)
        output += get_concept_arguments(self.right)
        return output

    @property
    def row_arguments(self) -> List[BuildConcept]:
        output = []
        output += get_concept_row_arguments(self.left)
        output += get_concept_row_arguments(self.right)
        return output

    @property
    def existence_arguments(self) -> list[tuple[BuildConcept, ...]]:
        output: list[tuple[BuildConcept, ...]] = []
        if isinstance(self.left, BuildConceptArgs):
            output += self.left.existence_arguments
        if isinstance(self.right, BuildConceptArgs):
            output += self.right.existence_arguments
        return output

    def decompose(self):
        chunks = []
        if self.operator == BooleanOperator.AND:
            for val in [self.left, self.right]:
                if isinstance(val, BuildConditional):
                    chunks.extend(val.decompose())
                else:
                    chunks.append(val)
        else:
            chunks.append(self)
        return chunks


@dataclass
class BuildWhereClause(BuildConceptArgs):
    conditional: Union[
        BuildSubselectComparison,
        BuildComparison,
        BuildConditional,
        BuildParenthetical,
    ]

    def __eq__(self, other):
        if not isinstance(other, (BuildWhereClause, WhereClause)):
            return False
        return self.conditional == other.conditional

    def __repr__(self):
        return str(self.conditional)

    def __str__(self):
        return self.__repr__()

    @property
    def concept_arguments(self) -> List[BuildConcept]:
        return self.conditional.concept_arguments

    @property
    def row_arguments(self) -> Sequence[BuildConcept]:
        return self.conditional.row_arguments

    @property
    def existence_arguments(self) -> Sequence[tuple["BuildConcept", ...]]:
        return self.conditional.existence_arguments


class BuildHavingClause(BuildWhereClause):
    pass


@dataclass
class BuildComparison(BuildConceptArgs, ConstantInlineable):

    left: Union[
        int,
        str,
        float,
        bool,
        datetime,
        date,
        BuildFunction,
        BuildConcept,
        BuildConditional,
        DataType,
        BuildComparison,
        BuildParenthetical,
        MagicConstants,
        BuildWindowItem,
        BuildAggregateWrapper,
        ListWrapper,
        TupleWrapper,
    ]
    right: Union[
        int,
        str,
        float,
        bool,
        date,
        datetime,
        BuildConcept,
        BuildFunction,
        BuildConditional,
        DataType,
        BuildComparison,
        BuildParenthetical,
        MagicConstants,
        BuildWindowItem,
        BuildAggregateWrapper,
        TupleWrapper,
        ListWrapper,
    ]
    operator: ComparisonOperator

    def __add__(self, other):
        if other is None:
            return self
        if not isinstance(
            other, (BuildComparison, BuildConditional, BuildParenthetical)
        ):
            raise ValueError(f"Cannot add {type(other)} to {__class__}")
        if other == self:
            return self
        return BuildConditional(left=self, right=other, operator=BooleanOperator.AND)

    def __repr__(self):
        if isinstance(self.left, BuildConcept):
            left = self.left.address
        else:
            left = str(self.left)
        if isinstance(self.right, BuildConcept):
            right = self.right.address
        else:
            right = str(self.right)
        return f"{left} {self.operator.value} {right}"

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if not isinstance(other, BuildComparison):
            return False
        return (
            self.left == other.left
            and self.right == other.right
            and self.operator == other.operator
        )

    def inline_constant(self, constant: BuildConcept):
        assert isinstance(constant.lineage, BuildFunction)
        new_val = constant.lineage.arguments[0]
        if isinstance(self.left, ConstantInlineable):
            new_left = self.left.inline_constant(constant)
        elif (
            isinstance(self.left, BuildConcept)
            and self.left.address == constant.address
        ):
            new_left = new_val
        else:
            new_left = self.left
        if isinstance(self.right, ConstantInlineable):
            new_right = self.right.inline_constant(constant)
        elif (
            isinstance(self.right, BuildConcept)
            and self.right.address == constant.address
        ):
            new_right = new_val
        else:
            new_right = self.right

        return self.__class__(
            left=new_left,
            right=new_right,
            operator=self.operator,
        )

    @property
    def concept_arguments(self) -> List[BuildConcept]:
        """Return BuildConcepts directly referenced in where clause"""
        output = []
        output += get_concept_arguments(self.left)
        output += get_concept_arguments(self.right)
        return output

    @property
    def row_arguments(self) -> List[BuildConcept]:
        output = []
        output += get_concept_row_arguments(self.left)
        output += get_concept_row_arguments(self.right)
        return output

    @property
    def existence_arguments(self) -> List[Tuple[BuildConcept, ...]]:
        """Return BuildConcepts directly referenced in where clause"""
        output: List[Tuple[BuildConcept, ...]] = []
        if isinstance(self.left, BuildConceptArgs):
            output += self.left.existence_arguments
        if isinstance(self.right, BuildConceptArgs):
            output += self.right.existence_arguments
        return output


@dataclass
class BuildSubselectComparison(BuildComparison):
    left: Union[
        int,
        str,
        float,
        bool,
        datetime,
        date,
        BuildFunction,
        BuildConcept,
        "BuildConditional",
        DataType,
        "BuildComparison",
        "BuildParenthetical",
        MagicConstants,
        BuildWindowItem,
        BuildAggregateWrapper,
        ListWrapper,
        TupleWrapper,
    ]
    right: Union[
        int,
        str,
        float,
        bool,
        date,
        datetime,
        BuildConcept,
        BuildFunction,
        "BuildConditional",
        DataType,
        "BuildComparison",
        "BuildParenthetical",
        MagicConstants,
        BuildWindowItem,
        BuildAggregateWrapper,
        TupleWrapper,
        ListWrapper,
    ]
    operator: ComparisonOperator

    def __eq__(self, other):
        if not isinstance(other, BuildSubselectComparison):
            return False

        comp = (
            self.left == other.left
            and self.right == other.right
            and self.operator == other.operator
        )
        return comp

    @property
    def row_arguments(self) -> List[BuildConcept]:
        return get_concept_row_arguments(self.left)

    @property
    def existence_arguments(self) -> list[tuple["BuildConcept", ...]]:
        return [tuple(get_concept_arguments(self.right))]


@dataclass
class BuildConcept(Addressable, BuildConceptArgs, DataTyped):
    model_config = ConfigDict(extra="forbid")
    name: str
    datatype: DataType | ArrayType | StructType | MapType | NumericType | TraitDataType
    purpose: Purpose
    build_is_aggregate: bool
    derivation: Derivation = Derivation.ROOT
    granularity: Granularity = Granularity.MULTI_ROW
    namespace: Optional[str] = field(default=DEFAULT_NAMESPACE)
    metadata: Metadata = field(
        default_factory=lambda: Metadata(description=None, line_number=None),
    )
    lineage: Optional[
        Union[
            BuildFunction,
            BuildWindowItem,
            BuildFilterItem,
            BuildAggregateWrapper,
            BuildRowsetItem,
            BuildMultiSelectLineage,
        ]
    ] = None

    keys: Optional[set[str]] = None
    grain: BuildGrain = field(default=None)  # type: ignore
    modifiers: List[Modifier] = field(default_factory=list)  # type: ignore
    pseudonyms: set[str] = field(default_factory=set)

    @property
    def is_aggregate(self) -> bool:
        return self.build_is_aggregate

    @cached_property
    def hash(self) -> int:
        return hash(
            f"{self.name}+{self.datatype}+ {self.purpose} + {str(self.lineage)} + {self.namespace} + {str(self.grain)} + {str(self.keys)}"
        )

    def __hash__(self):
        return self.hash

    def __repr__(self):
        base = f"{self.address}@{self.grain}"
        return base

    @property
    def output_datatype(self):
        return self.datatype

    def __eq__(self, other: object):
        if isinstance(other, str):
            if self.address == other:
                return True
        if not isinstance(other, (BuildConcept, Concept)):
            return False
        return (
            self.name == other.name
            and self.datatype == other.datatype
            and self.purpose == other.purpose
            and self.namespace == other.namespace
            and self.grain == other.grain
            # and self.keys == other.keys
        )

    def __str__(self):
        grain = str(self.grain) if self.grain else "Grain<>"
        return f"{self.namespace}.{self.name}@{grain}"

    @cached_property
    def address(self) -> str:
        return f"{self.namespace}.{self.name}"

    @property
    def output(self) -> "BuildConcept":
        return self

    @property
    def safe_address(self) -> str:
        if self.namespace == DEFAULT_NAMESPACE:
            return self.name.replace(".", "_")
        elif self.namespace:
            return f"{self.namespace.replace('.','_')}_{self.name.replace('.','_')}"
        return self.name.replace(".", "_")

    def with_grain(self, grain: Optional["BuildGrain" | BuildConcept] = None) -> Self:
        if isinstance(grain, BuildConcept):
            grain = BuildGrain(
                components=set(
                    [
                        grain.address,
                    ]
                )
            )
        return self.__class__(
            name=self.name,
            datatype=self.datatype,
            purpose=self.purpose,
            metadata=self.metadata,
            lineage=self.lineage,
            grain=grain if grain else BuildGrain(components=set()),
            namespace=self.namespace,
            keys=self.keys,
            modifiers=self.modifiers,
            pseudonyms=self.pseudonyms,
            ## bound
            derivation=self.derivation,
            granularity=self.granularity,
            build_is_aggregate=self.build_is_aggregate,
        )

    @property
    def _with_default_grain(self) -> Self:
        if self.purpose == Purpose.KEY:
            # we need to make this abstract
            grain = BuildGrain(components={self.address})
        elif self.purpose == Purpose.PROPERTY:
            components = []
            if self.keys:
                components = [*self.keys]
            if self.lineage:
                for item in self.lineage.concept_arguments:
                    components += [x.address for x in item.sources]
            # TODO: set synonyms
            grain = BuildGrain(
                components=set([x for x in components]),
            )  # synonym_set=generate_BuildConcept_synonyms(components))
        elif self.purpose == Purpose.METRIC:
            grain = BuildGrain()
        elif self.purpose == Purpose.CONSTANT:
            if self.derivation != Derivation.CONSTANT:
                grain = BuildGrain(components={self.address})
            else:
                grain = self.grain
        else:
            grain = self.grain  # type: ignore
        return self.__class__(
            name=self.name,
            datatype=self.datatype,
            purpose=self.purpose,
            metadata=self.metadata,
            lineage=self.lineage,
            grain=grain,
            keys=self.keys,
            namespace=self.namespace,
            modifiers=self.modifiers,
            pseudonyms=self.pseudonyms,
            ## bound
            derivation=self.derivation,
            granularity=self.granularity,
            build_is_aggregate=self.build_is_aggregate,
        )

    def with_default_grain(self) -> "BuildConcept":
        return self._with_default_grain

    @property
    def sources(self) -> List["BuildConcept"]:
        if self.lineage:
            output: List[BuildConcept] = []

            def get_sources(
                expr: Union[
                    BuildFunction,
                    BuildWindowItem,
                    BuildFilterItem,
                    BuildAggregateWrapper,
                    BuildRowsetItem,
                    BuildMultiSelectLineage,
                ],
                output: List[BuildConcept],
            ):
                if isinstance(expr, BuildMultiSelectLineage):
                    return
                for item in expr.concept_arguments:
                    if isinstance(item, BuildConcept):
                        if item.address == self.address:
                            raise SyntaxError(
                                f"BuildConcept {self.address} references itself"
                            )
                        output.append(item)
                        output += item.sources

            get_sources(self.lineage, output)
            return output
        return []

    @property
    def concept_arguments(self) -> List[BuildConcept]:
        return self.lineage.concept_arguments if self.lineage else []


@dataclass
class BuildOrderItem(DataTyped, BuildConceptArgs):
    expr: BuildExpr
    order: Ordering

    @property
    def concept_arguments(self) -> List[BuildConcept]:
        base: List[BuildConcept] = []
        x = self.expr
        if isinstance(x, BuildConcept):
            base += [x]
        elif isinstance(x, BuildConceptArgs):
            base += x.concept_arguments
        return base

    @property
    def row_arguments(self) -> Sequence[BuildConcept]:
        if isinstance(self.expr, BuildConceptArgs):
            return self.expr.row_arguments
        return self.concept_arguments

    @property
    def existence_arguments(self) -> Sequence[tuple["BuildConcept", ...]]:
        if isinstance(self.expr, BuildConceptArgs):
            return self.expr.existence_arguments
        return []

    @property
    def output_datatype(self):
        return arg_to_datatype(self.expr)


@dataclass
class BuildWindowItem(DataTyped, BuildConceptArgs):
    type: WindowType
    content: BuildConcept
    order_by: List[BuildOrderItem]
    over: List["BuildConcept"] = field(default_factory=list)
    index: Optional[int] = None

    def __repr__(self) -> str:
        return f"{self.type}({self.content} {self.index}, {self.over}, {self.order_by})"

    def __str__(self):
        return self.__repr__()

    @property
    def concept_arguments(self) -> List[BuildConcept]:
        output = [self.content]
        for order in self.order_by:
            output += order.concept_arguments
        for item in self.over:
            output += [item]
        return output

    @property
    def output(self) -> BuildConcept:
        return self.content

    @property
    def output_datatype(self):
        if self.type in (WindowType.RANK, WindowType.ROW_NUMBER):
            return DataType.INTEGER
        return self.content.output_datatype

    @property
    def output_purpose(self):
        return Purpose.PROPERTY


@dataclass
class BuildCaseWhen(BuildConceptArgs):
    comparison: BuildConditional | BuildSubselectComparison | BuildComparison
    expr: "BuildExpr"

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"WHEN {str(self.comparison)} THEN {str(self.expr)}"

    @property
    def concept_arguments(self):
        return get_concept_arguments(self.comparison) + get_concept_arguments(self.expr)

    @property
    def concept_row_arguments(self):
        return get_concept_row_arguments(self.comparison) + get_concept_row_arguments(
            self.expr
        )


@dataclass
class BuildCaseElse(BuildConceptArgs):
    expr: "BuildExpr"

    @property
    def concept_arguments(self):
        return get_concept_arguments(self.expr)


@dataclass
class BuildFunction(DataTyped, BuildConceptArgs):
    operator: FunctionType
    arguments: Sequence[
        Union[
            int,
            float,
            str,
            date,
            datetime,
            MapWrapper[Any, Any],
            TraitDataType,
            DataType,
            ArrayType,
            MapType,
            NumericType,
            DatePart,
            BuildConcept,
            BuildAggregateWrapper,
            BuildFunction,
            BuildWindowItem,
            BuildParenthetical,
            BuildCaseWhen,
            BuildCaseElse,
            list,
            ListWrapper[Any],
        ]
    ]
    output_data_type: (
        DataType | ArrayType | StructType | MapType | NumericType | TraitDataType
    )
    output_purpose: Purpose = field(default=Purpose.KEY)
    arg_count: int = field(default=1)
    valid_inputs: Optional[
        Union[
            Set[DataType],
            List[Set[DataType]],
        ]
    ] = None

    def __repr__(self):
        return f'{self.operator.value}({",".join([str(a) for a in self.arguments])})'

    def __str__(self):
        return self.__repr__()

    @property
    def datatype(self):
        return self.output_data_type

    @property
    def output_datatype(self):
        return self.output_data_type

    @property
    def concept_arguments(self) -> List[BuildConcept]:
        base = []
        for arg in self.arguments:
            base += get_concept_arguments(arg)
        return base

    @property
    def output_grain(self):
        # aggregates have an abstract grain
        if self.operator in FunctionClass.AGGREGATE_FUNCTIONS.value:
            return BuildGrain(components=[])
        # scalars have implicit grain of all arguments
        args = set()
        for input in self.concept_arguments:
            args += input.grain.components
        return BuildGrain(components=args)


@dataclass
class BuildAggregateWrapper(BuildConceptArgs, DataTyped):
    function: BuildFunction
    by: List[BuildConcept] = field(default_factory=list)

    def __str__(self):
        grain_str = [str(c) for c in self.by] if self.by else "abstract"
        return f"{str(self.function)}<{grain_str}>"

    @property
    def datatype(self):
        return self.function.datatype

    @property
    def concept_arguments(self) -> List[BuildConcept]:
        return self.function.concept_arguments + self.by

    @property
    def output_datatype(self):
        return self.function.output_datatype

    @property
    def output_purpose(self):
        return self.function.output_purpose


@dataclass
class BuildFilterItem(BuildConceptArgs):
    content: "BuildExpr"
    where: BuildWhereClause

    def __str__(self):
        return f"<Filter: {str(self.content)} where {str(self.where)}>"

    @property
    def output_datatype(self):
        return arg_to_datatype(self.content)

    @property
    def output_purpose(self):
        return self.content.purpose

    @property
    def content_concept_arguments(self):
        if isinstance(self.content, BuildConcept):
            return [self.content]
        elif isinstance(self.content, BuildConceptArgs):
            return self.content.concept_arguments
        return []

    @property
    def concept_arguments(self):
        if isinstance(self.content, BuildConcept):
            return [self.content] + self.where.concept_arguments
        elif isinstance(self.content, BuildConceptArgs):
            return self.content.concept_arguments + self.where.concept_arguments
        return self.where.concept_arguments


@dataclass
class BuildRowsetLineage(BuildConceptArgs):
    name: str
    derived_concepts: List[str]
    select: SelectLineage | MultiSelectLineage


@dataclass
class BuildRowsetItem(DataTyped, BuildConceptArgs):
    content: BuildConcept
    rowset: BuildRowsetLineage

    def __repr__(self):
        return f"<Rowset<{self.rowset.name}>: {str(self.content)}>"

    def __str__(self):
        return self.__repr__()

    @property
    def output(self) -> BuildConcept:
        return self.content

    @property
    def output_datatype(self):
        return self.content.datatype

    @property
    def output_purpose(self):
        return self.content.purpose

    @property
    def concept_arguments(self):
        return [self.content]


@dataclass
class BuildOrderBy:
    items: List[BuildOrderItem]

    @property
    def concept_arguments(self):
        return [x.expr for x in self.items]


@dataclass
class BuildAlignClause:
    items: List[BuildAlignItem]


@dataclass
class BuildSelectLineage:
    selection: List[BuildConcept]
    hidden_components: set[str]
    local_concepts: dict[str, BuildConcept]
    order_by: Optional[BuildOrderBy] = None
    limit: Optional[int] = None
    meta: Metadata = field(default_factory=lambda: Metadata())
    grain: BuildGrain = field(default_factory=BuildGrain)
    where_clause: BuildWhereClause | None = field(default=None)
    having_clause: BuildHavingClause | None = field(default=None)

    @property
    def output_components(self) -> List[BuildConcept]:
        return self.selection


@dataclass
class BuildMultiSelectLineage(BuildConceptArgs):
    selects: List[SelectLineage]
    grain: BuildGrain
    align: BuildAlignClause
    namespace: str
    local_concepts: dict[str, BuildConcept]
    build_concept_arguments: list[BuildConcept]
    build_output_components: list[BuildConcept]
    hidden_components: set[str]
    order_by: Optional[BuildOrderBy] = None
    limit: Optional[int] = None
    where_clause: Union["BuildWhereClause", None] = field(default=None)
    having_clause: Union["BuildHavingClause", None] = field(default=None)

    @property
    def derived_concepts(self) -> set[str]:
        output = set()
        for item in self.align.items:
            output.add(item.aligned_concept)
        return output

    @property
    def output_components(self) -> list[BuildConcept]:
        return self.build_output_components

    @property
    def derived_concept(self) -> set[str]:
        output = set()
        for item in self.align.items:
            output.add(item.aligned_concept)
        return output

    @property
    def concept_arguments(self) -> list[BuildConcept]:
        return self.build_concept_arguments

    # these are needed to help disambiguate between parents
    def get_merge_concept(self, check: BuildConcept) -> str | None:
        for item in self.align.items:
            if check in item.concepts_lcl:
                return f"{item.namespace}.{item.alias}"
        return None

    def find_source(self, concept: BuildConcept, cte: CTE | UnionCTE) -> BuildConcept:
        for x in self.align.items:
            if concept.name == x.alias:
                for c in x.concepts:
                    if c.address in cte.output_lcl:
                        return c
        raise SyntaxError(
            f"Could not find upstream map for multiselect {str(concept)} on cte ({cte})"
        )


@dataclass
class BuildAlignItem:
    alias: str
    concepts: List[BuildConcept]
    namespace: str = field(default=DEFAULT_NAMESPACE)

    @cached_property
    def concepts_lcl(self) -> LooseBuildConceptList:
        return LooseBuildConceptList(concepts=self.concepts)

    @property
    def aligned_concept(self) -> str:
        return f"{self.namespace}.{self.alias}"


@dataclass
class BuildColumnAssignment:
    alias: str | RawColumnExpr | BuildFunction | BuildAggregateWrapper
    concept: BuildConcept
    modifiers: List[Modifier] = field(default_factory=list)

    @property
    def is_complete(self) -> bool:
        return Modifier.PARTIAL not in self.modifiers

    @property
    def is_nullable(self) -> bool:
        return Modifier.NULLABLE in self.modifiers


@dataclass
class BuildDatasource:
    name: str
    columns: List[BuildColumnAssignment]
    address: Union[Address, str]
    grain: BuildGrain = field(
        default_factory=lambda: BuildGrain(components=set()),
    )
    namespace: Optional[str] = field(default=DEFAULT_NAMESPACE)
    metadata: DatasourceMetadata = field(
        default_factory=lambda: DatasourceMetadata(freshness_concept=None)
    )
    where: Optional[BuildWhereClause] = None
    non_partial_for: Optional[BuildWhereClause] = None

    def __hash__(self):
        return self.identifier.__hash__()

    def __add__(self, other):
        if not other == self:
            raise ValueError(
                "Attempted to add two datasources that are not identical, this is not a valid operation"
            )
        return self

    @property
    def condition(self):
        return None

    @property
    def can_be_inlined(self) -> bool:
        if isinstance(self.address, Address) and self.address.is_query:
            return False
        return True

    @property
    def identifier(self) -> str:
        if not self.namespace or self.namespace == DEFAULT_NAMESPACE:
            return self.name
        return f"{self.namespace}.{self.name}"

    @property
    def safe_identifier(self) -> str:
        return self.identifier.replace(".", "_")

    @property
    def concepts(self) -> List[BuildConcept]:
        return [c.concept for c in self.columns]

    @property
    def group_required(self):
        return False

    @property
    def output_concepts(self) -> List[BuildConcept]:
        return self.concepts

    @property
    def full_concepts(self) -> List[BuildConcept]:
        return [c.concept for c in self.columns if Modifier.PARTIAL not in c.modifiers]

    @property
    def nullable_concepts(self) -> List[BuildConcept]:
        return [c.concept for c in self.columns if Modifier.NULLABLE in c.modifiers]

    @property
    def hidden_concepts(self) -> List[BuildConcept]:
        return [c.concept for c in self.columns if Modifier.HIDDEN in c.modifiers]

    @property
    def partial_concepts(self) -> List[BuildConcept]:
        return [c.concept for c in self.columns if Modifier.PARTIAL in c.modifiers]

    def get_alias(
        self,
        concept: BuildConcept,
        use_raw_name: bool = True,
        force_alias: bool = False,
    ) -> Optional[str | RawColumnExpr] | BuildFunction | BuildAggregateWrapper:
        # 2022-01-22
        # this logic needs to be refined.
        # if concept.lineage:
        # #     return None
        for x in self.columns:
            if x.concept == concept or x.concept.with_grain(concept.grain) == concept:
                if use_raw_name:
                    return x.alias
                return concept.safe_address
        existing = [str(c.concept) for c in self.columns]
        raise ValueError(
            f"{LOGGER_PREFIX} Concept {concept} not found on {self.identifier}; have"
            f" {existing}."
        )

    @property
    def safe_location(self) -> str:
        if isinstance(self.address, Address):
            return self.address.location
        return self.address

    @property
    def output_lcl(self) -> LooseBuildConceptList:
        return LooseBuildConceptList(concepts=self.output_concepts)


BuildExpr = (
    BuildWindowItem
    | BuildFilterItem
    | BuildConcept
    | BuildComparison
    | BuildConditional
    | BuildParenthetical
    | BuildFunction
    | BuildAggregateWrapper
    | bool
    | MagicConstants
    | int
    | str
    | float
    | list
)


def get_canonical_pseudonyms(environment: Environment) -> dict[str, set[str]]:
    roots: dict[str, set[str]] = defaultdict(set)
    for k, v in environment.concepts.items():
        roots[v.address].add(k)
        for x in v.pseudonyms:
            roots[v.address].add(x)
    for k, v in environment.alias_origin_lookup.items():
        lookup = environment.concepts[k].address
        roots[lookup].add(v.address)
        for x2 in v.pseudonyms:
            roots[lookup].add(x2)
    return roots


def requires_concept_nesting(
    expr,
) -> AggregateWrapper | WindowItem | FilterItem | Function | None:
    if isinstance(expr, (AggregateWrapper, WindowItem, FilterItem)):
        return expr
    if isinstance(expr, Function) and expr.operator in (
        FunctionType.GROUP,
        FunctionType.PARENTHETICAL,
    ):
        # group by requires nesting
        return expr
    return None


class Factory:

    def __init__(
        self,
        environment: Environment,
        local_concepts: dict[str, BuildConcept] | None = None,
        grain: Grain | None = None,
        pseudonym_map: dict[str, set[str]] | None = None,
    ):
        self.grain = grain or Grain()
        self.environment = environment
        self.local_concepts: dict[str, BuildConcept] = (
            {} if local_concepts is None else local_concepts
        )
        self.local_non_build_concepts: dict[str, Concept] = {}
        self.pseudonym_map = pseudonym_map or get_canonical_pseudonyms(environment)
        self.build_grain = self.build(self.grain) if self.grain else None

    def instantiate_concept(
        self,
        arg: (
            AggregateWrapper
            | FunctionCallWrapper
            | WindowItem
            | FilterItem
            | Function
            | ListWrapper
            | MapWrapper
            | int
            | float
            | str
            | date
        ),
    ) -> tuple[Concept, BuildConcept]:
        from trilogy.parsing.common import arbitrary_to_concept, generate_concept_name

        name = generate_concept_name(arg)
        if name in self.local_concepts and name in self.local_non_build_concepts:
            # if we already have this concept, return it
            return self.local_non_build_concepts[name], self.local_concepts[name]
        new = arbitrary_to_concept(
            arg,
            environment=self.environment,
        )
        built = self._build_concept(new)
        self.local_concepts[name] = built
        self.local_non_build_concepts[name] = new
        return new, built

    @singledispatchmethod
    def build(self, base):
        raise NotImplementedError("Cannot build {}".format(type(base)))

    @build.register
    def _(
        self,
        base: (
            int
            | str
            | float
            | list
            | date
            | TupleWrapper
            | ListWrapper
            | MagicConstants
            | MapWrapper
            | DataType
            | DatePart
            | NumericType
        ),
    ) -> (
        int
        | str
        | float
        | list
        | date
        | TupleWrapper
        | ListWrapper
        | MagicConstants
        | MapWrapper
        | DataType
        | DatePart
        | NumericType
    ):
        return self._build_primitive(base)

    def _build_primitive(self, base):
        return base

    @build.register
    def _(self, base: None) -> None:
        return self._build_none(base)

    def _build_none(self, base):
        return base

    @build.register
    def _(self, base: Function) -> BuildFunction | BuildAggregateWrapper:
        return self._build_function(base)

    def _build_function(self, base: Function) -> BuildFunction | BuildAggregateWrapper:
        raw_args: list[Concept | FuncArgs] = []
        for arg in base.arguments:
            # to do proper discovery, we need to inject virtual intermediate concepts
            # we don't use requires_concept_nesting here by design
            if isinstance(arg, (AggregateWrapper, FilterItem, WindowItem)):
                narg, _ = self.instantiate_concept(arg)
                raw_args.append(narg)
            else:
                raw_args.append(arg)
        if base.operator == FunctionType.GROUP:
            group_base = raw_args[0]
            final_args: List[Concept | ConceptRef] = []
            if isinstance(group_base, ConceptRef):
                if group_base.address in self.environment.concepts and not isinstance(
                    self.environment.concepts[group_base.address], UndefinedConcept
                ):
                    group_base = self.environment.concepts[group_base.address]
            if (
                isinstance(group_base, Concept)
                and isinstance(group_base.lineage, AggregateWrapper)
                and not group_base.lineage.by
            ):
                arguments = raw_args[1:]
                for x in arguments:
                    if isinstance(x, (ConceptRef, Concept)):
                        final_args.append(x)
                    else:
                        # constants, etc, can be ignored for group
                        continue
                _, rval = self.instantiate_concept(
                    AggregateWrapper(
                        function=group_base.lineage.function,
                        by=final_args,
                    )
                )

                return BuildFunction(
                    operator=base.operator,
                    arguments=[
                        rval,
                        *[self.handle_constant(self.build(c)) for c in raw_args[1:]],
                    ],
                    output_data_type=base.output_datatype,
                    output_purpose=base.output_purpose,
                    valid_inputs=base.valid_inputs,
                    arg_count=base.arg_count,
                )

        new = BuildFunction(
            operator=base.operator,
            arguments=[self.handle_constant(self.build(c)) for c in raw_args],
            output_data_type=base.output_datatype,
            output_purpose=base.output_purpose,
            valid_inputs=base.valid_inputs,
            arg_count=base.arg_count,
        )
        return new

    @build.register
    def _(self, base: ConceptRef) -> BuildConcept:
        return self._build_concept_ref(base)

    def _build_concept_ref(self, base: ConceptRef) -> BuildConcept:
        if base.address in self.local_concepts:
            full = self.local_concepts[base.address]
            if isinstance(full, BuildConcept):
                return full
        if base.address in self.environment.concepts:
            raw = self.environment.concepts[base.address]
            return self._build_concept(raw)
        # this will error by design - TODO - more helpful message?
        return self._build_concept(self.environment.concepts[base.address])

    @build.register
    def _(self, base: CaseWhen) -> BuildCaseWhen:
        return self._build_case_when(base)

    def _build_case_when(self, base: CaseWhen) -> BuildCaseWhen:
        expr: Concept | FuncArgs = base.expr
        validation = requires_concept_nesting(expr)
        if validation:
            expr, _ = self.instantiate_concept(validation)
        return BuildCaseWhen(
            comparison=self.build(base.comparison),
            expr=self.build(expr),
        )

    @build.register
    def _(self, base: CaseElse) -> BuildCaseElse:
        return self._build_case_else(base)

    def _build_case_else(self, base: CaseElse) -> BuildCaseElse:
        expr: Concept | FuncArgs = base.expr
        validation = requires_concept_nesting(expr)
        if validation:
            expr, _ = self.instantiate_concept(validation)
        return BuildCaseElse(expr=self.build(expr))

    @build.register
    def _(self, base: Concept) -> BuildConcept:
        return self._build_concept(base)

    def _build_concept(self, base: Concept) -> BuildConcept:
        # TODO: if we are using parameters, wrap it in a new model and use that in rendering
        if base.address in self.local_concepts:
            return self.local_concepts[base.address]
        new_lineage, final_grain, _ = base.get_select_grain_and_keys(
            self.grain, self.environment
        )
        if new_lineage:
            build_lineage = self.build(new_lineage)
        else:
            build_lineage = None
        derivation = Concept.calculate_derivation(build_lineage, base.purpose)

        granularity = Concept.calculate_granularity(
            derivation, final_grain, build_lineage
        )
        is_aggregate = Concept.calculate_is_aggregate(build_lineage)

        # if this is a pseudonym, we need to look up the base address
        if base.address in self.environment.alias_origin_lookup:
            lookup_address = self.environment.concepts[base.address].address
            # map only to the canonical concept, not to other merged concepts
            base_pseudonyms = {lookup_address}
        else:
            base_pseudonyms = {
                x
                for x in self.pseudonym_map.get(base.address, set())
                if x != base.address
            }
        rval = BuildConcept(
            name=base.name,
            datatype=base.datatype,
            purpose=base.purpose,
            metadata=base.metadata,
            lineage=build_lineage,
            grain=self._build_grain(final_grain),
            namespace=base.namespace,
            keys=base.keys,
            modifiers=base.modifiers,
            pseudonyms=base_pseudonyms,
            ## instantiated values
            derivation=derivation,
            granularity=granularity,
            build_is_aggregate=is_aggregate,
        )
        self.local_concepts[base.address] = rval
        return rval

    @build.register
    def _(self, base: AggregateWrapper) -> BuildAggregateWrapper:
        return self._build_aggregate_wrapper(base)

    def _build_aggregate_wrapper(self, base: AggregateWrapper) -> BuildAggregateWrapper:
        if not base.by:
            by = [
                self._build_concept(self.environment.concepts[c])
                for c in self.grain.components
            ]
        else:
            by = [self.build(x) for x in base.by]

        parent: BuildFunction = self._build_function(base.function)  # type: ignore
        return BuildAggregateWrapper(function=parent, by=by)

    @build.register
    def _(self, base: ColumnAssignment) -> BuildColumnAssignment:
        return self._build_column_assignment(base)

    def _build_column_assignment(self, base: ColumnAssignment) -> BuildColumnAssignment:
        address = base.concept.address
        fetched = (
            self._build_concept(
                self.environment.alias_origin_lookup[address]
            ).with_grain(self.build_grain)
            if address in self.environment.alias_origin_lookup
            else self._build_concept(self.environment.concepts[address]).with_grain(
                self.build_grain
            )
        )

        return BuildColumnAssignment(
            alias=(
                self._build_function(base.alias)
                if isinstance(base.alias, Function)
                else base.alias
            ),
            concept=fetched,
            modifiers=base.modifiers,
        )

    @build.register
    def _(self, base: OrderBy) -> BuildOrderBy:
        return self._build_order_by(base)

    def _build_order_by(self, base: OrderBy) -> BuildOrderBy:
        return BuildOrderBy(items=[self._build_order_item(x) for x in base.items])

    @build.register
    def _(self, base: FunctionCallWrapper) -> BuildExpr:
        return self._build_function_call_wrapper(base)

    def _build_function_call_wrapper(self, base: FunctionCallWrapper) -> BuildExpr:
        # function calls are kept around purely for the parse tree
        # so discard at the build point
        return self.build(base.content)

    @build.register
    def _(self, base: OrderItem) -> BuildOrderItem:
        return self._build_order_item(base)

    def _build_order_item(self, base: OrderItem) -> BuildOrderItem:
        bexpr: Any
        validation = requires_concept_nesting(base.expr)
        if validation:
            bexpr, _ = self.instantiate_concept(validation)
        else:
            bexpr = base.expr
        return BuildOrderItem(
            expr=(self.build(bexpr)),
            order=base.order,
        )

    @build.register
    def _(self, base: WhereClause) -> BuildWhereClause:
        return self._build_where_clause(base)

    def _build_where_clause(self, base: WhereClause) -> BuildWhereClause:
        return BuildWhereClause(conditional=self.build(base.conditional))

    @build.register
    def _(self, base: HavingClause) -> BuildHavingClause:
        return self._build_having_clause(base)

    def _build_having_clause(self, base: HavingClause) -> BuildHavingClause:
        return BuildHavingClause(conditional=self.build(base.conditional))

    @build.register
    def _(self, base: WindowItem) -> BuildWindowItem:
        return self._build_window_item(base)

    def _build_window_item(self, base: WindowItem) -> BuildWindowItem:
        content: Concept | FuncArgs = base.content
        validation = requires_concept_nesting(base.content)
        if validation:
            content, _ = self.instantiate_concept(validation)
        final_by = []
        for x in base.order_by:
            if (
                isinstance(x.expr, AggregateWrapper)
                and not x.expr.by
                and isinstance(content, (ConceptRef, Concept))
            ):
                x.expr.by = [content]
            final_by.append(x)
        return BuildWindowItem(
            type=base.type,
            content=self.build(content),
            order_by=[self.build(x) for x in final_by],
            over=[self._build_concept_ref(x) for x in base.over],
            index=base.index,
        )

    @build.register
    def _(self, base: Conditional) -> BuildConditional:
        return self._build_conditional(base)

    def _build_conditional(self, base: Conditional) -> BuildConditional:
        return BuildConditional(
            left=self.handle_constant(self.build(base.left)),
            right=self.handle_constant(self.build(base.right)),
            operator=base.operator,
        )

    @build.register
    def _(self, base: SubselectComparison) -> BuildSubselectComparison:
        return self._build_subselect_comparison(base)

    def _build_subselect_comparison(
        self, base: SubselectComparison
    ) -> BuildSubselectComparison:
        right: Any = base.right
        # this has specialized logic - include all Functions
        if isinstance(base.right, (AggregateWrapper, WindowItem, FilterItem, Function)):
            right_c, _ = self.instantiate_concept(base.right)
            right = right_c
        return BuildSubselectComparison(
            left=self.handle_constant(self.build(base.left)),
            right=self.handle_constant(self.build(right)),
            operator=base.operator,
        )

    @build.register
    def _(self, base: Comparison) -> BuildComparison:
        return self._build_comparison(base)

    def _build_comparison(self, base: Comparison) -> BuildComparison:
        left = base.left
        validation = requires_concept_nesting(base.left)
        if validation:
            left_c, _ = self.instantiate_concept(validation)
            left = left_c  # type: ignore
        right = base.right
        validation = requires_concept_nesting(base.right)
        if validation:
            right_c, _ = self.instantiate_concept(validation)
            right = right_c  # type: ignore
        return BuildComparison(
            left=self.handle_constant(self.build(left)),
            right=self.handle_constant(self.build(right)),
            operator=base.operator,
        )

    @build.register
    def _(self, base: AlignItem) -> BuildAlignItem:
        return self._build_align_item(base)

    def _build_align_item(self, base: AlignItem) -> BuildAlignItem:
        return BuildAlignItem(
            alias=base.alias,
            concepts=[self._build_concept_ref(x) for x in base.concepts],
            namespace=base.namespace,
        )

    @build.register
    def _(self, base: AlignClause) -> BuildAlignClause:
        return self._build_align_clause(base)

    def _build_align_clause(self, base: AlignClause) -> BuildAlignClause:
        return BuildAlignClause(items=[self._build_align_item(x) for x in base.items])

    @build.register
    def _(self, base: RowsetItem) -> BuildRowsetItem:
        return self._build_rowset_item(base)

    def _build_rowset_item(self, base: RowsetItem) -> BuildRowsetItem:
        factory = Factory(
            environment=self.environment,
            local_concepts={},
            grain=base.rowset.select.grain,
            pseudonym_map=self.pseudonym_map,
        )
        return BuildRowsetItem(
            content=factory._build_concept_ref(base.content),
            rowset=factory._build_rowset_lineage(base.rowset),
        )

    @build.register
    def _(self, base: RowsetLineage) -> BuildRowsetLineage:
        return self._build_rowset_lineage(base)

    def _build_rowset_lineage(self, base: RowsetLineage) -> BuildRowsetLineage:
        out = BuildRowsetLineage(
            name=base.name,
            derived_concepts=[x.address for x in base.derived_concepts],
            select=base.select,
        )
        return out

    @build.register
    def _(self, base: Grain) -> BuildGrain:
        return self._build_grain(base)

    def _build_grain(self, base: Grain) -> BuildGrain:
        if base.where_clause:
            factory = Factory(
                environment=self.environment, pseudonym_map=self.pseudonym_map
            )
            where = factory._build_where_clause(base.where_clause)
        else:
            where = None
        return BuildGrain(components=base.components, where_clause=where)

    @build.register
    def _(self, base: TupleWrapper) -> TupleWrapper:
        return self._build_tuple_wrapper(base)

    def _build_tuple_wrapper(self, base: TupleWrapper) -> TupleWrapper:
        return TupleWrapper(val=[self.build(x) for x in base.val], type=base.type)

    @build.register
    def _(self, base: FilterItem) -> BuildFilterItem:
        return self._build_filter_item(base)

    def _build_filter_item(self, base: FilterItem) -> BuildFilterItem:
        if isinstance(
            base.content, (Function, AggregateWrapper, WindowItem, FilterItem)
        ):
            _, built = self.instantiate_concept(base.content)
            return BuildFilterItem(content=built, where=self.build(base.where))
        return BuildFilterItem(
            content=self.build(base.content), where=self.build(base.where)
        )

    @build.register
    def _(self, base: Parenthetical) -> BuildParenthetical:
        return self._build_parenthetical(base)

    def _build_parenthetical(self, base: Parenthetical) -> BuildParenthetical:
        validate = requires_concept_nesting(base.content)
        if validate:
            content, _ = self.instantiate_concept(validate)
            return BuildParenthetical(content=self.build(content))
        else:
            return BuildParenthetical(content=self.build(base.content))

    @build.register
    def _(self, base: SelectLineage) -> BuildSelectLineage:
        return self._build_select_lineage(base)

    def _build_select_lineage(self, base: SelectLineage) -> BuildSelectLineage:
        from trilogy.core.models.build import (
            BuildSelectLineage,
            Factory,
        )

        materialized: dict[str, BuildConcept] = {}
        factory = Factory(
            grain=base.grain,
            environment=self.environment,
            local_concepts=materialized,
            pseudonym_map=self.pseudonym_map,
        )
        for k, v in base.local_concepts.items():
            materialized[k] = factory.build(v)
        where_factory = Factory(
            grain=Grain(),
            environment=self.environment,
            local_concepts={},
            pseudonym_map=self.pseudonym_map,
        )
        where_clause = (
            where_factory.build(base.where_clause) if base.where_clause else None
        )
        # if the where clause derives new concepts
        # we need to ensure these are accessible from the general factory
        # post resolution
        for bk, bv in where_factory.local_concepts.items():
            # but do not override any local cahced grains
            if bk in materialized:
                continue
            materialized[bk] = bv
        final: List[BuildConcept] = []
        for original in base.selection:
            new = original
            # we don't know the grain of an aggregate at assignment time
            # so rebuild at this point in the tree
            # TODO: simplify
            if new.address in materialized:
                built = materialized[new.address]
            else:
                # Sometimes cached values here don't have the latest info
                # but we can't just use environment, as it might not have the right grain.
                built = factory.build(new)
                materialized[new.address] = built
            final.append(built)
        return BuildSelectLineage(
            selection=final,
            hidden_components=base.hidden_components,
            order_by=(factory.build(base.order_by) if base.order_by else None),
            limit=base.limit,
            meta=base.meta,
            local_concepts=materialized,
            grain=self.build(base.grain),
            having_clause=(
                factory.build(base.having_clause) if base.having_clause else None
            ),
            # this uses a different grain factory
            where_clause=where_clause,
        )

    @build.register
    def _(self, base: MultiSelectLineage) -> BuildMultiSelectLineage:
        return self._build_multi_select_lineage(base)

    def _build_multi_select_lineage(
        self, base: MultiSelectLineage
    ) -> BuildMultiSelectLineage:
        local_build_cache: dict[str, BuildConcept] = {}

        parents: list[BuildSelectLineage] = [self.build(x) for x in base.selects]
        base_local = parents[0].local_concepts

        for select in parents[1:]:
            for k, v in select.local_concepts.items():
                base_local[k] = v

        # this requires custom handling to avoid circular dependencies
        final_grain = self.build(base.grain)
        derived_base = []
        for k in base.derived_concepts:
            base_concept = self.environment.concepts[k]
            x = BuildConcept(
                name=base_concept.name,
                datatype=base_concept.datatype,
                purpose=base_concept.purpose,
                build_is_aggregate=False,
                derivation=Derivation.MULTISELECT,
                lineage=None,
                grain=final_grain,
                namespace=base_concept.namespace,
            )
            local_build_cache[k] = x
            derived_base.append(x)
        all_input: list[BuildConcept] = []
        for parent in parents:
            all_input += parent.output_components
        all_output: list[BuildConcept] = derived_base + all_input
        final: list[BuildConcept] = [
            x for x in all_output if x.address not in base.hidden_components
        ]
        factory = Factory(
            grain=base.grain,
            environment=self.environment,
            local_concepts=local_build_cache,
            pseudonym_map=self.pseudonym_map,
        )
        where_factory = Factory(
            environment=self.environment, pseudonym_map=self.pseudonym_map
        )
        lineage = BuildMultiSelectLineage(
            # we don't build selects here; they'll be built automatically in query discovery
            selects=base.selects,
            grain=final_grain,
            align=factory.build(base.align),
            # self.align.with_select_context(
            #     local_build_cache, self.grain, environment
            # ),
            namespace=base.namespace,
            hidden_components=base.hidden_components,
            order_by=factory.build(base.order_by) if base.order_by else None,
            limit=base.limit,
            where_clause=(
                where_factory.build(base.where_clause) if base.where_clause else None
            ),
            having_clause=(
                factory.build(base.having_clause) if base.having_clause else None
            ),
            local_concepts=base_local,
            build_output_components=final,
            build_concept_arguments=all_input,
        )
        for k in base.derived_concepts:
            local_build_cache[k].lineage = lineage
        return lineage

    @build.register
    def _(self, base: Environment):
        return self._build_environment(base)

    def _build_environment(self, base: Environment):
        from trilogy.core.models.build_environment import BuildEnvironment

        new = BuildEnvironment(
            namespace=base.namespace,
            cte_name_map=base.cte_name_map,
        )

        for k, v in base.concepts.items():
            new.concepts[k] = self._build_concept(v)
        for (
            k,
            d,
        ) in base.datasources.items():
            new.datasources[k] = self._build_datasource(d)
        for k, a in base.alias_origin_lookup.items():
            new.alias_origin_lookup[k] = self._build_concept(a)
        # add in anything that was built as a side-effect
        for bk, bv in self.local_concepts.items():
            if bk not in new.concepts:
                new.concepts[bk] = bv
        new.gen_concept_list_caches()
        return new

    @build.register
    def _(self, base: TraitDataType):
        return self._build_trait_data_type(base)

    def _build_trait_data_type(self, base: TraitDataType):
        return base

    @build.register
    def _(self, base: ArrayType):
        return self._build_array_type(base)

    def _build_array_type(self, base: ArrayType):
        return base

    @build.register
    def _(self, base: StructType):
        return self._build_struct_type(base)

    def _build_struct_type(self, base: StructType):
        return base

    @build.register
    def _(self, base: MapType):
        return self._build_map_type(base)

    def _build_map_type(self, base: MapType):
        return base

    @build.register
    def _(self, base: ArgBinding):
        return self._build_arg_binding(base)

    def _build_arg_binding(self, base: ArgBinding):
        return base

    @build.register
    def _(self, base: Ordering):
        return self._build_ordering(base)

    def _build_ordering(self, base: Ordering):
        return base

    @build.register
    def _(self, base: Datasource):
        return self._build_datasource(base)

    def _build_datasource(self, base: Datasource):
        local_cache: dict[str, BuildConcept] = {}
        factory = Factory(
            grain=base.grain,
            environment=self.environment,
            local_concepts=local_cache,
            pseudonym_map=self.pseudonym_map,
        )
        return BuildDatasource(
            name=base.name,
            columns=[factory._build_column_assignment(c) for c in base.columns],
            address=base.address,
            grain=factory._build_grain(base.grain),
            namespace=base.namespace,
            metadata=base.metadata,
            where=(factory.build(base.where) if base.where else None),
            non_partial_for=(
                factory.build(base.non_partial_for) if base.non_partial_for else None
            ),
        )

    def handle_constant(self, base):
        if (
            isinstance(base, BuildConcept)
            and isinstance(base.lineage, BuildFunction)
            and base.lineage.operator == FunctionType.CONSTANT
        ):
            return BuildParamaterizedConceptReference(concept=base)
        elif isinstance(base, ConceptRef):
            return self.handle_constant(self.build(base))
        return base



from typing import List, Optional, Tuple

from trilogy.constants import logger
from trilogy.core.enums import (
    JoinType,
    SourceType,
)
from trilogy.core.models.build import (
    BuildComparison,
    BuildConcept,
    BuildConditional,
    BuildDatasource,
    BuildGrain,
    BuildOrderBy,
    BuildParenthetical,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.models.execute import BaseJoin, QueryDatasource, UnnestJoin
from trilogy.core.processing.nodes.base_node import (
    NodeJoin,
    StrategyNode,
    resolve_concept_map,
)
from trilogy.core.processing.utility import find_nullable_concepts, get_node_joins
from trilogy.utility import unique

LOGGER_PREFIX = "[CONCEPT DETAIL - MERGE NODE]"


def deduplicate_nodes(
    merged: dict[str, QueryDatasource | BuildDatasource],
    logging_prefix: str,
    environment: BuildEnvironment,
) -> tuple[bool, dict[str, QueryDatasource | BuildDatasource], set[str]]:
    duplicates = False
    removed: set[str] = set()
    set_map: dict[str, set[str]] = {}
    for k, v in merged.items():
        unique_outputs = [
            # the concept may be a in a different environment for a rowset.
            (environment.concepts.get(x.address) or x).address
            for x in v.output_concepts
            if x not in v.partial_concepts
        ]
        set_map[k] = set(unique_outputs)
    for k1, v1 in set_map.items():
        found = False
        for k2, v2 in set_map.items():
            if k1 == k2:
                continue
            if (
                v1.issubset(v2)
                and merged[k1].grain.issubset(merged[k2].grain)
                and not merged[k2].partial_concepts
                and not merged[k1].partial_concepts
                and not merged[k2].condition
                and not merged[k1].condition
            ):
                og = merged[k1]
                subset_to = merged[k2]
                logger.info(
                    f"{logging_prefix}{LOGGER_PREFIX} extraneous parent node that is subset of another parent node {og.grain.issubset(subset_to.grain)} {og.grain.components} {subset_to.grain.components}"
                )
                merged = {k: v for k, v in merged.items() if k != k1}
                removed.add(k1)
                duplicates = True
                found = True
                break
        if found:
            break

    return duplicates, merged, removed


def deduplicate_nodes_and_joins(
    joins: List[NodeJoin] | None,
    merged: dict[str, QueryDatasource | BuildDatasource],
    logging_prefix: str,
    environment: BuildEnvironment,
) -> Tuple[List[NodeJoin] | None, dict[str, QueryDatasource | BuildDatasource]]:
    # it's possible that we have more sources than we need
    duplicates = True
    while duplicates:
        duplicates = False
        duplicates, merged, removed = deduplicate_nodes(
            merged, logging_prefix, environment=environment
        )
        # filter out any removed joins
        if joins is not None:
            joins = [
                j
                for j in joins
                if j.left_node.resolve().identifier not in removed
                and j.right_node.resolve().identifier not in removed
            ]
    return joins, merged


class MergeNode(StrategyNode):
    source_type = SourceType.MERGE

    def __init__(
        self,
        input_concepts: List[BuildConcept],
        output_concepts: List[BuildConcept],
        environment,
        whole_grain: bool = False,
        parents: List["StrategyNode"] | None = None,
        node_joins: List[NodeJoin] | None = None,
        join_concepts: Optional[List] = None,
        force_join_type: Optional[JoinType] = None,
        partial_concepts: Optional[List[BuildConcept]] = None,
        nullable_concepts: Optional[List[BuildConcept]] = None,
        force_group: bool | None = None,
        depth: int = 0,
        grain: BuildGrain | None = None,
        conditions: (
            BuildConditional | BuildComparison | BuildParenthetical | None
        ) = None,
        preexisting_conditions: (
            BuildConditional | BuildComparison | BuildParenthetical | None
        ) = None,
        hidden_concepts: set[str] | None = None,
        virtual_output_concepts: List[BuildConcept] | None = None,
        existence_concepts: List[BuildConcept] | None = None,
        ordering: BuildOrderBy | None = None,
    ):
        super().__init__(
            input_concepts=input_concepts,
            output_concepts=output_concepts,
            environment=environment,
            whole_grain=whole_grain,
            parents=parents,
            depth=depth,
            partial_concepts=partial_concepts,
            nullable_concepts=nullable_concepts,
            force_group=force_group,
            grain=grain,
            conditions=conditions,
            preexisting_conditions=preexisting_conditions,
            hidden_concepts=hidden_concepts,
            virtual_output_concepts=virtual_output_concepts,
            existence_concepts=existence_concepts,
            ordering=ordering,
        )
        self.join_concepts = join_concepts
        self.force_join_type = force_join_type
        self.node_joins: List[NodeJoin] | None = node_joins

        final_joins: List[NodeJoin] = []
        if self.node_joins is not None:
            for join in self.node_joins:
                if join.left_node.resolve().name == join.right_node.resolve().name:
                    continue
                final_joins.append(join)
            self.node_joins = final_joins

    def translate_node_joins(self, node_joins: List[NodeJoin]) -> List[BaseJoin]:
        joins = []
        for join in node_joins:
            left = join.left_node.resolve()
            right = join.right_node.resolve()
            if left.identifier == right.identifier:
                raise SyntaxError(f"Cannot join node {left.identifier} to itself")
            joins.append(
                BaseJoin(
                    left_datasource=left,
                    right_datasource=right,
                    join_type=join.join_type,
                    concepts=join.concepts,
                    concept_pairs=join.concept_pairs,
                )
            )
        return joins

    def create_full_joins(self, dataset_list: List[QueryDatasource | BuildDatasource]):
        joins = []
        seen = set()
        for left_value in dataset_list:
            for right_value in dataset_list:
                if left_value.identifier == right_value.identifier:
                    continue
                if left_value.identifier in seen and right_value.identifier in seen:
                    continue
                joins.append(
                    BaseJoin(
                        left_datasource=left_value,
                        right_datasource=right_value,
                        join_type=JoinType.FULL,
                        concepts=[],
                    )
                )
                seen.add(left_value.identifier)
                seen.add(right_value.identifier)
        return joins

    def generate_joins(
        self,
        final_datasets,
        final_joins: List[NodeJoin] | None,
        pregrain: BuildGrain,
        grain: BuildGrain,
        environment: BuildEnvironment,
    ) -> List[BaseJoin | UnnestJoin]:
        # only finally, join between them for unique values
        dataset_list: List[QueryDatasource | BuildDatasource] = sorted(
            final_datasets, key=lambda x: -len(x.grain.components)
        )

        logger.info(
            f"{self.logging_prefix}{LOGGER_PREFIX} Merge node has {len(dataset_list)} parents, starting merge"
        )
        if final_joins is None:
            if not pregrain.components:
                logger.info(
                    f"{self.logging_prefix}{LOGGER_PREFIX} no grain components, doing full join"
                )
                joins = self.create_full_joins(dataset_list)
            else:
                logger.info(
                    f"{self.logging_prefix}{LOGGER_PREFIX} inferring node joins to target grain {str(grain)}"
                )
                joins = get_node_joins(dataset_list, environment=environment)
        elif final_joins:
            logger.info(
                f"{self.logging_prefix}{LOGGER_PREFIX} translating provided node joins {len(final_joins)}"
            )
            joins = self.translate_node_joins(final_joins)
        else:
            logger.info(
                f"{self.logging_prefix}{LOGGER_PREFIX} Final joins is not null {final_joins} but is empty, skipping join generation"
            )
            return []
        return joins

    def _resolve(self) -> QueryDatasource:
        parent_sources: List[QueryDatasource | BuildDatasource] = [
            p.resolve() for p in self.parents
        ]
        merged: dict[str, QueryDatasource | BuildDatasource] = {}
        final_joins: List[NodeJoin] | None = self.node_joins
        for source in parent_sources:
            if source.identifier in merged:
                logger.info(
                    f"{self.logging_prefix}{LOGGER_PREFIX} merging parent node with {source.identifier} into existing"
                )
                merged[source.identifier] = merged[source.identifier] + source
            else:
                merged[source.identifier] = source

        # it's possible that we have more sources than we need
        final_joins, merged = deduplicate_nodes_and_joins(
            final_joins, merged, self.logging_prefix, self.environment
        )
        # early exit if we can just return the parent
        final_datasets: List[QueryDatasource | BuildDatasource] = list(merged.values())

        existence_final = [
            x
            for x in final_datasets
            if all([y in self.existence_concepts for y in x.output_concepts])
        ]
        if len(merged.keys()) == 1:
            final: QueryDatasource | BuildDatasource = list(merged.values())[0]
            if (
                set([c.address for c in final.output_concepts])
                == set([c.address for c in self.output_concepts])
                and not self.conditions
                and isinstance(final, QueryDatasource)
            ):
                logger.info(
                    f"{self.logging_prefix}{LOGGER_PREFIX} Merge node has only one parent with the same"
                    " outputs as this merge node, dropping merge node "
                )
                # push up any conditions we need
                final.ordering = self.ordering
                return final

        # if we have multiple candidates, see if one is good enough
        for dataset in final_datasets:
            output_set = set(
                [
                    c.address
                    for c in dataset.output_concepts
                    if c.address not in [x.address for x in dataset.partial_concepts]
                ]
            )
            if (
                all([c.address in output_set for c in self.all_concepts])
                and not self.conditions
                and isinstance(dataset, QueryDatasource)
            ):
                logger.info(
                    f"{self.logging_prefix}{LOGGER_PREFIX} Merge node not required as parent node {dataset.source_type}"
                    f" has all required output properties with partial {[c.address for c in dataset.partial_concepts]}"
                    f" and self has no conditions ({self.conditions})"
                )
                dataset.ordering = self.ordering
                return dataset

        pregrain = BuildGrain()

        for source in final_datasets:
            if all(
                [x.address in self.existence_concepts for x in source.output_concepts]
            ):
                logger.info(
                    f"{self.logging_prefix}{LOGGER_PREFIX} skipping existence only source with {source.output_concepts} from grain accumulation"
                )
                continue
            logger.info(
                f"{self.logging_prefix}{LOGGER_PREFIX} adding source grain {source.grain} from source {source.identifier} to pregrain"
            )
            pregrain += source.grain
            logger.info(
                f"{self.logging_prefix}{LOGGER_PREFIX} pregrain is now {pregrain}"
            )

        pregrain = BuildGrain.from_concepts(
            pregrain.components, environment=self.environment
        )

        grain = self.grain if self.grain else pregrain
        logger.info(
            f"{self.logging_prefix}{LOGGER_PREFIX} has pre grain {pregrain} and final merge node grain {grain}"
        )
        join_candidates = [x for x in final_datasets if x not in existence_final]
        if len(join_candidates) > 1:
            joins: List[BaseJoin | UnnestJoin] = self.generate_joins(
                join_candidates, final_joins, pregrain, grain, self.environment
            )
        else:
            joins = []

        logger.info(
            f"{self.logging_prefix}{LOGGER_PREFIX} Final join count for CTE parent count {len(join_candidates)} is {len(joins)}"
        )
        full_join_concepts = []
        for join in joins:
            if isinstance(join, BaseJoin) and join.join_type == JoinType.FULL:
                full_join_concepts += join.input_concepts

        if self.force_group is True:

            force_group = True
        elif self.whole_grain:
            force_group = False
        elif self.force_group is False:
            force_group = False
        elif not any(
            [d.grain.issubset(grain) for d in final_datasets]
        ) and not pregrain.issubset(grain):
            logger.info(
                f"{self.logging_prefix}{LOGGER_PREFIX} no parents include full grain {grain} and pregrain {pregrain} does not match, assume must group to grain. Have {[str(d.grain) for d in final_datasets]}"
            )
            force_group = True
        else:
            force_group = None

        qd_joins: List[BaseJoin | UnnestJoin] = [*joins]

        source_map = resolve_concept_map(
            final_datasets,
            targets=self.output_concepts,
            inherited_inputs=self.input_concepts + self.existence_concepts,
            full_joins=full_join_concepts,
        )
        nullable_concepts = find_nullable_concepts(
            source_map=source_map, joins=joins, datasources=final_datasets
        )
        if force_group:

            grain = BuildGrain.from_concepts(
                self.output_concepts, environment=self.environment
            )
            logger.info(
                f"{self.logging_prefix}{LOGGER_PREFIX} forcing group by to achieve grain {grain}"
            )
        qds = QueryDatasource(
            input_concepts=unique(self.input_concepts, "address"),
            output_concepts=unique(self.output_concepts, "address"),
            datasources=final_datasets,
            source_type=self.source_type,
            source_map=source_map,
            joins=qd_joins,
            grain=grain,
            nullable_concepts=[
                x for x in self.output_concepts if x.address in nullable_concepts
            ],
            partial_concepts=self.partial_concepts,
            force_group=force_group,
            condition=self.conditions,
            hidden_concepts=self.hidden_concepts,
            ordering=self.ordering,
        )
        return qds

    def copy(self) -> "MergeNode":
        return MergeNode(
            input_concepts=list(self.input_concepts),
            output_concepts=list(self.output_concepts),
            environment=self.environment,
            whole_grain=self.whole_grain,
            parents=self.parents,
            depth=self.depth,
            partial_concepts=list(self.partial_concepts),
            force_group=self.force_group,
            grain=self.grain,
            conditions=self.conditions,
            preexisting_conditions=self.preexisting_conditions,
            nullable_concepts=list(self.nullable_concepts),
            hidden_concepts=set(self.hidden_concepts),
            virtual_output_concepts=list(self.virtual_output_concepts),
            node_joins=list(self.node_joins) if self.node_joins else None,
            join_concepts=list(self.join_concepts) if self.join_concepts else None,
            force_join_type=self.force_join_type,
            existence_concepts=list(self.existence_concepts),
            ordering=self.ordering,
        )



# from trilogy.compiler import compile
from trilogy.core.models.author import Grain
from trilogy.core.query_processor import process_query
from trilogy.core.statements.author import SelectStatement
from trilogy.parser import parse

TEST_SETUP = r"""


key order_key int;
property order_key.order_amount float;
metric total_order_amount <- sum(order_amount);

key key int;
property key.weight float;
property key.name string;


key sub_category_key int;
key category_key int;

property category_key.category_name string;


datasource product_info (
productKey:key,
productSubCategoryKey:sub_category_key,
englishProductName:name
)
grain (key)
address AdventureWorksDW2019.dbo.DimProduct
;

datasource orders (
order:order_key,
product_key:key,
order_amount:order_amount
)
grain(order_key)
address AdventureWorksDW2019.dbo.DimOrder;


datasource product_sub_category (
productSubCategoryKey:sub_category_key,
productCategoryKey:category_key,
)
grain (sub_category_key)
address AdventureWorksDW2019.dbo.DimProductSubCategory
;

datasource product_category (
productCategoryKey:category_key,
englishProductCategoryName:category_name
)
address AdventureWorksDW2019.dbo.DimProductCategory
;

select 
    category_name,
    total_order_amount;
"""


def test_select():
    env, parsed = parse(TEST_SETUP)
    select: SelectStatement = parsed[-1]

    assert select.grain == Grain(components=[env.concepts["category_name"]])

    process_query(statement=select, environment=env)



from dataclasses import dataclass, field
from typing import List, Optional, Union

from trilogy.core.enums import IOType, ValidationScope
from trilogy.core.models.author import ConceptRef, HavingClause, WhereClause
from trilogy.core.models.build import (
    BuildConcept,
    BuildDatasource,
    BuildOrderBy,
)
from trilogy.core.models.datasource import Address, Datasource
from trilogy.core.models.environment import EnvironmentConceptDict
from trilogy.core.models.execute import CTE, UnionCTE


@dataclass
class CopyQueryMixin:
    target: str
    target_type: IOType


@dataclass
class MaterializedDataset:
    address: Address


@dataclass
class PersistQueryMixin:
    output_to: MaterializedDataset
    datasource: Datasource


@dataclass
class SelectTypeMixin:
    where_clause: Union["WhereClause", None] = field(default=None)
    having_clause: Union["HavingClause", None] = field(default=None)

    @property
    def output_components(self) -> List[ConceptRef]:
        raise NotImplementedError


@dataclass
class ProcessedQuery:
    output_columns: List[ConceptRef]
    ctes: List[CTE | UnionCTE]
    base: CTE | UnionCTE
    hidden_columns: set[str] = field(default_factory=set)
    limit: Optional[int] = None
    order_by: Optional[BuildOrderBy] = None
    local_concepts: EnvironmentConceptDict = field(
        default_factory=EnvironmentConceptDict
    )
    locally_derived: set[str] = field(default_factory=set)


@dataclass
class ProcessedQueryPersist(ProcessedQuery, PersistQueryMixin):
    pass


@dataclass
class ProcessedCopyStatement(ProcessedQuery, CopyQueryMixin):
    pass


@dataclass
class ProcessedRawSQLStatement:
    text: str


@dataclass
class ProcessedValidateStatement:
    scope: ValidationScope
    targets: Optional[List[str]]


@dataclass
class ProcessedStaticValueOutput:
    values: List[dict]


@dataclass
class ProcessedShowStatement:
    output_columns: List[ConceptRef]
    output_values: List[
        Union[
            BuildConcept,
            BuildDatasource,
            ProcessedQuery,
            ProcessedQueryPersist,
            ProcessedCopyStatement,
            ProcessedValidateStatement,
            ProcessedStaticValueOutput,
        ]
    ]


PROCESSED_STATEMENT_TYPES = (
    ProcessedCopyStatement
    | ProcessedQuery
    | ProcessedRawSQLStatement
    | ProcessedQueryPersist
    | ProcessedShowStatement
    | ProcessedValidateStatement
)



from enum import Enum
from typing import TYPE_CHECKING, Callable, List, Optional

from trilogy.core.models.environment import Environment

if TYPE_CHECKING:
    from trilogy import Executor
    from trilogy.hooks.base_hook import BaseHook

from trilogy.constants import Rendering, logger
from trilogy.dialect.config import DialectConfig


def default_factory(conf: DialectConfig, config_type):
    from sqlalchemy import create_engine

    if not isinstance(conf, config_type):
        raise TypeError(
            f"Invalid dialect configuration for type {type(config_type).__name__}, is {type(conf)}"
        )
    if conf.connect_args:
        return create_engine(
            conf.connection_string(), future=True, connect_args=conf.connect_args
        )
    return create_engine(conf.connection_string(), future=True)


class Dialects(Enum):
    BIGQUERY = "bigquery"
    SQL_SERVER = "sql_server"
    DUCK_DB = "duck_db"
    PRESTO = "presto"
    TRINO = "trino"
    POSTGRES = "postgres"
    SNOWFLAKE = "snowflake"
    DATAFRAME = "dataframe"

    @classmethod
    def _missing_(cls, value):
        if value == "duckdb":
            return cls.DUCK_DB
        return super()._missing_(value)

    def default_renderer(self, conf=None, _engine_factory: Callable = default_factory):
        from trilogy.render import get_dialect_generator

        return get_dialect_generator(self)

    def default_engine(self, conf=None, _engine_factory: Callable = default_factory):
        if self == Dialects.BIGQUERY:
            from google.auth import default
            from google.cloud import bigquery

            from trilogy.dialect.config import BigQueryConfig

            credentials, project = default()
            client = bigquery.Client(credentials=credentials, project=project)
            conf = conf or BigQueryConfig(project=project, client=client)
            return _engine_factory(
                conf,
                BigQueryConfig,
            )
        elif self == Dialects.SQL_SERVER:
            raise NotImplementedError()
        elif self == Dialects.DUCK_DB:
            from trilogy.dialect.config import DuckDBConfig

            if not conf:
                conf = DuckDBConfig()
            return _engine_factory(conf, DuckDBConfig)
        elif self == Dialects.SNOWFLAKE:
            from trilogy.dialect.config import SnowflakeConfig

            return _engine_factory(conf, SnowflakeConfig)
        elif self == Dialects.POSTGRES:
            logger.warn(
                "WARN: Using experimental postgres dialect. Most functionality will not work."
            )
            import importlib

            spec = importlib.util.find_spec("psycopg2")
            if spec is None:
                raise ImportError(
                    "postgres driver not installed. python -m pip install pypreql[postgres]"
                )
            from trilogy.dialect.config import PostgresConfig

            return _engine_factory(conf, PostgresConfig)
        elif self == Dialects.PRESTO:
            from trilogy.dialect.config import PrestoConfig

            return _engine_factory(conf, PrestoConfig)
        elif self == Dialects.TRINO:
            from trilogy.dialect.config import TrinoConfig

            return _engine_factory(conf, TrinoConfig)
        elif self == Dialects.DATAFRAME:
            from trilogy.dialect.config import DataFrameConfig
            from trilogy.dialect.dataframe import DataframeConnectionWrapper

            if not conf:
                conf = DataFrameConfig(dataframes={})

            base = _engine_factory(conf, DataFrameConfig)

            return DataframeConnectionWrapper(base, dataframes=conf.dataframes)
        else:
            raise ValueError(
                f"Unsupported dialect {self} for default engine creation; create one explicitly."
            )

    def default_executor(
        self,
        environment: Optional["Environment"] = None,
        hooks: List["BaseHook"] | None = None,
        conf: DialectConfig | None = None,
        rendering: Rendering | None = None,
        _engine_factory: Callable | None = None,
    ) -> "Executor":
        from trilogy import Executor

        if _engine_factory is not None:
            return Executor(
                engine=self.default_engine(conf=conf, _engine_factory=_engine_factory),
                environment=environment or Environment(),
                dialect=self,
                rendering=rendering,
                hooks=hooks,
            )

        return Executor(
            engine=self.default_engine(conf=conf),
            environment=environment or Environment(),
            dialect=self,
            rendering=rendering,
            hooks=hooks,
        )



from dataclasses import dataclass
from datetime import date, datetime
from typing import Any, Callable, Optional

from lark.tree import Meta

from trilogy.constants import MagicConstants
from trilogy.core.enums import (
    DatePart,
    FunctionClass,
    FunctionType,
    Granularity,
    InfiniteFunctionArgs,
    Purpose,
)
from trilogy.core.exceptions import InvalidSyntaxException
from trilogy.core.models.author import (
    AggregateWrapper,
    Concept,
    ConceptRef,
    Conditional,
    Function,
    Parenthetical,
    UndefinedConcept,
    WindowItem,
)
from trilogy.core.models.core import (
    CONCRETE_TYPES,
    ArrayType,
    DataType,
    MapType,
    NumericType,
    StructType,
    TraitDataType,
    arg_to_datatype,
    merge_datatypes,
)
from trilogy.core.models.environment import Environment

GENERIC_ARGS = Concept | ConceptRef | Function | str | int | float | date | datetime


@dataclass
class FunctionConfig:
    arg_count: int = 1
    valid_inputs: set[DataType] | list[set[DataType]] | None = None
    output_purpose: Purpose | None = None
    output_type: (
        DataType | ArrayType | MapType | StructType | NumericType | TraitDataType | None
    ) = None
    output_type_function: Optional[Callable] = None


def get_unnest_output_type(args: list[Any]) -> CONCRETE_TYPES:
    output = arg_to_datatype(args[0])
    if isinstance(output, (ArrayType, MapType)):
        output = output.value_data_type
    else:
        output = DataType.STRING
    return output


def get_coalesce_output_type(args: list[Any]) -> CONCRETE_TYPES:
    non_null = [x for x in args if not x == MagicConstants.NULL]
    processed = [arg_to_datatype(x) for x in non_null if x]
    if not len(set(processed)) == 1:
        raise InvalidSyntaxException(
            f"All arguments to coalesce must be of the same type, have {set(arg_to_datatype(x) for x in args)}"
        )
    return processed[0]


def get_transform_output_type(args: list[Any]) -> CONCRETE_TYPES:
    return ArrayType(type=arg_to_datatype(args[2]))


def get_index_output_type(
    args: list[Any],
) -> CONCRETE_TYPES:
    arg = args[0]
    datatype = arg_to_datatype(arg)
    if isinstance(datatype, ArrayType):
        return datatype.value_data_type
    elif isinstance(datatype, MapType):
        return datatype.value_data_type
    return datatype


def get_attr_datatype(
    args: list[Any],
) -> CONCRETE_TYPES:
    arg = args[0]
    lookup = args[1]
    datatype = arg_to_datatype(arg)
    if isinstance(datatype, StructType):
        return datatype.field_types[lookup]
    return datatype


def get_cast_output_type(
    args: list[Any],
) -> DataType:
    return args[1]


def get_output_type_at_index(args, index: int):
    return arg_to_datatype(args[index])


def validate_case_output(
    args: list[Any],
) -> DataType:
    datatypes = set()
    mapz = dict()
    for arg in args:
        output_datatype = arg_to_datatype(arg.expr)
        if output_datatype != DataType.NULL:
            datatypes.add(output_datatype.data_type)
        mapz[str(arg.expr)] = output_datatype
    known = [x for x in datatypes if x != DataType.UNKNOWN]
    if len(known) == 0:
        return DataType.UNKNOWN
    if not len(known) == 1:
        raise SyntaxError(
            f"All case expressions must have the same output datatype, got {datatypes} from {mapz}"
        )
    return known.pop()


def create_struct_output(
    args: list[Any],
) -> StructType:
    zipped = dict(zip(args[1::2], args[::2]))
    types = [arg_to_datatype(x) for x in args[::2]]
    return StructType(fields=types, fields_map=zipped)


def get_date_part_output(args: list[Any]):
    target = args[1]
    if target == DatePart.YEAR:
        return TraitDataType(type=DataType.INTEGER, traits=["year"])
    elif target == DatePart.MONTH:
        return TraitDataType(type=DataType.INTEGER, traits=["month"])
    elif target == DatePart.DAY:
        return TraitDataType(type=DataType.INTEGER, traits=["day"])
    elif target == DatePart.HOUR:
        return TraitDataType(type=DataType.INTEGER, traits=["hour"])
    elif target == DatePart.MINUTE:
        return TraitDataType(type=DataType.INTEGER, traits=["minute"])
    elif target == DatePart.SECOND:
        return TraitDataType(type=DataType.INTEGER, traits=["second"])
    elif target == DatePart.WEEK:
        return TraitDataType(type=DataType.INTEGER, traits=["week"])
    elif target == DatePart.QUARTER:
        return TraitDataType(type=DataType.INTEGER, traits=["quarter"])
    elif target == DatePart.DAY_OF_WEEK:
        return TraitDataType(type=DataType.INTEGER, traits=["day_of_week"])
    else:
        raise InvalidSyntaxException(f"Date part not supported for {target}")


def get_date_trunc_output(
    args: list[Any],
):
    target: DatePart = args[1]
    if target == DatePart.YEAR:
        return DataType.DATE
    elif target == DatePart.MONTH:
        return DataType.DATE
    elif target == DatePart.DAY:
        return DataType.DATE
    elif target == DatePart.HOUR:
        return DataType.DATETIME
    elif target == DatePart.MINUTE:
        return DataType.DATETIME
    elif target == DatePart.SECOND:
        return DataType.DATETIME
    else:
        raise InvalidSyntaxException(f"Date truncation not supported for {target}")


def get_map_key_type(arg):
    arg_datatype = arg_to_datatype(arg)
    if isinstance(arg_datatype, MapType):
        return ArrayType(type=arg_datatype.key_data_type)
    return ArrayType(type=DataType.STRING)


def get_map_value_type(arg):
    arg_datatype = arg_to_datatype(arg)
    if isinstance(arg_datatype, MapType):
        return ArrayType(type=arg_datatype.value_data_type)
    return ArrayType(type=DataType.STRING)


FUNCTION_REGISTRY: dict[FunctionType, FunctionConfig] = {
    FunctionType.ALIAS: FunctionConfig(
        arg_count=1,
    ),
    FunctionType.PARENTHETICAL: FunctionConfig(
        arg_count=1,
    ),
    FunctionType.UNNEST: FunctionConfig(
        valid_inputs={
            DataType.ARRAY,
        },
        output_purpose=Purpose.KEY,
        output_type_function=get_unnest_output_type,
        arg_count=1,
    ),
    FunctionType.RECURSE_EDGE: FunctionConfig(
        arg_count=2,
    ),
    FunctionType.GROUP: FunctionConfig(
        arg_count=-1,
        output_type_function=lambda args: get_output_type_at_index(args, 0),
    ),
    FunctionType.COUNT: FunctionConfig(
        output_purpose=Purpose.METRIC,
        output_type=DataType.INTEGER,
        arg_count=1,
    ),
    FunctionType.COUNT_DISTINCT: FunctionConfig(
        output_purpose=Purpose.METRIC,
        output_type=DataType.INTEGER,
        arg_count=1,
    ),
    FunctionType.MAX: FunctionConfig(
        valid_inputs={
            DataType.INTEGER,
            DataType.FLOAT,
            DataType.NUMBER,
            DataType.DATE,
            DataType.DATETIME,
            DataType.TIMESTAMP,
            DataType.BOOL,
        },
        output_purpose=Purpose.METRIC,
        arg_count=1,
    ),
    FunctionType.MIN: FunctionConfig(
        valid_inputs={
            DataType.INTEGER,
            DataType.FLOAT,
            DataType.NUMBER,
            DataType.DATE,
            DataType.DATETIME,
            DataType.TIMESTAMP,
        },
        output_purpose=Purpose.METRIC,
        arg_count=1,
    ),
    FunctionType.SPLIT: FunctionConfig(
        valid_inputs={DataType.STRING},
        output_purpose=Purpose.PROPERTY,
        output_type=ArrayType(type=DataType.STRING),
        arg_count=2,
    ),
    FunctionType.INDEX_ACCESS: FunctionConfig(
        valid_inputs=[
            {
                DataType.ARRAY,
            },
            {
                DataType.INTEGER,
            },
        ],
        output_purpose=Purpose.PROPERTY,
        output_type_function=get_index_output_type,
        arg_count=2,
    ),
    FunctionType.MAP_KEYS: FunctionConfig(
        valid_inputs={
            DataType.MAP,
        },
        output_purpose=Purpose.PROPERTY,
        output_type_function=lambda args: get_map_key_type(args[0]),
        arg_count=1,
    ),
    FunctionType.MAP_VALUES: FunctionConfig(
        valid_inputs={
            DataType.MAP,
        },
        output_purpose=Purpose.PROPERTY,
        output_type_function=lambda args: get_map_value_type(args[0]),
        arg_count=1,
    ),
    FunctionType.ARRAY_DISTINCT: FunctionConfig(
        valid_inputs={
            DataType.ARRAY,
        },
        output_purpose=Purpose.PROPERTY,
        output_type_function=lambda args: get_output_type_at_index(args, 0),
        arg_count=1,
    ),
    FunctionType.ARRAY_SORT: FunctionConfig(
        valid_inputs=[
            {DataType.ARRAY},
            {DataType.STRING},
        ],
        output_purpose=Purpose.PROPERTY,
        output_type_function=lambda args: get_output_type_at_index(args, 0),
        arg_count=2,
    ),
    FunctionType.ARRAY_TRANSFORM: FunctionConfig(
        valid_inputs=[
            {
                DataType.ARRAY,
            },
            {*DataType},
            {*DataType},
        ],
        output_purpose=Purpose.PROPERTY,
        output_type_function=get_transform_output_type,
        arg_count=3,
    ),
    FunctionType.ARRAY_TO_STRING: FunctionConfig(
        valid_inputs={
            DataType.ARRAY,
            DataType.STRING,
        },
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.STRING,
        arg_count=2,
    ),
    FunctionType.ARRAY_SUM: FunctionConfig(
        valid_inputs={
            DataType.ARRAY,
        },
        output_purpose=Purpose.PROPERTY,
        output_type_function=get_index_output_type,
        arg_count=1,
    ),
    FunctionType.MAP_ACCESS: FunctionConfig(
        valid_inputs=[
            {
                DataType.MAP,
            },
            {
                DataType.INTEGER,
                DataType.STRING,
            },
        ],
        output_purpose=Purpose.PROPERTY,
        output_type_function=get_index_output_type,
        arg_count=2,
    ),
    FunctionType.ATTR_ACCESS: FunctionConfig(
        valid_inputs=[
            {DataType.STRUCT},
            {
                DataType.STRING,
            },
        ],
        output_purpose=Purpose.PROPERTY,
        output_type_function=get_attr_datatype,
        arg_count=2,
    ),
    FunctionType.ABS: FunctionConfig(
        valid_inputs={DataType.INTEGER, DataType.FLOAT, DataType.NUMBER},
        output_purpose=Purpose.PROPERTY,
        arg_count=1,
    ),
    FunctionType.NULLIF: FunctionConfig(
        valid_inputs={*DataType},
        output_purpose=Purpose.PROPERTY,
        output_type_function=lambda args: get_output_type_at_index(args, 0),
        arg_count=2,
    ),
    FunctionType.COALESCE: FunctionConfig(
        valid_inputs={*DataType},
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.INTEGER,
        arg_count=-1,
        output_type_function=get_coalesce_output_type,
    ),
    FunctionType.CURRENT_DATE: FunctionConfig(
        output_purpose=Purpose.CONSTANT,
        output_type=DataType.DATE,
        arg_count=0,
    ),
    FunctionType.CURRENT_DATETIME: FunctionConfig(
        output_purpose=Purpose.CONSTANT,
        output_type=DataType.DATETIME,
        arg_count=0,
    ),
    FunctionType.CURRENT_TIMESTAMP: FunctionConfig(
        output_purpose=Purpose.CONSTANT,
        output_type=DataType.TIMESTAMP,
        arg_count=0,
    ),
    FunctionType.BOOL: FunctionConfig(
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.BOOL,
        arg_count=1,
    ),
    FunctionType.STRPOS: FunctionConfig(
        valid_inputs=[
            {DataType.STRING},
            {DataType.STRING},
        ],
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.INTEGER,
        arg_count=2,
    ),
    FunctionType.REPLACE: FunctionConfig(
        valid_inputs=[
            {DataType.STRING},
            {DataType.STRING},
            {DataType.STRING},
        ],
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.STRING,
        arg_count=3,
    ),
    FunctionType.CONTAINS: FunctionConfig(
        valid_inputs=[
            {DataType.STRING},
            {DataType.STRING},
        ],
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.BOOL,
        arg_count=2,
    ),
    FunctionType.TRIM: FunctionConfig(
        valid_inputs=[
            {DataType.STRING},
            {DataType.STRING},
        ],
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.STRING,
        arg_count=2,
    ),
    FunctionType.SUBSTRING: FunctionConfig(
        valid_inputs=[{DataType.STRING}, {DataType.INTEGER}, {DataType.INTEGER}],
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.STRING,
        arg_count=3,
    ),
    FunctionType.UNION: FunctionConfig(
        valid_inputs={*DataType},
        output_purpose=Purpose.KEY,
        arg_count=-1,
    ),
    FunctionType.LIKE: FunctionConfig(
        valid_inputs={DataType.STRING},
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.BOOL,
        arg_count=2,
    ),
    FunctionType.ILIKE: FunctionConfig(
        valid_inputs={DataType.STRING},
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.BOOL,
        arg_count=2,
    ),
    FunctionType.UPPER: FunctionConfig(
        valid_inputs={DataType.STRING},
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.STRING,
        arg_count=1,
    ),
    FunctionType.LOWER: FunctionConfig(
        valid_inputs={DataType.STRING},
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.STRING,
        arg_count=1,
    ),
    FunctionType.REGEXP_CONTAINS: FunctionConfig(
        valid_inputs={DataType.STRING},
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.BOOL,
        arg_count=2,
    ),
    FunctionType.REGEXP_EXTRACT: FunctionConfig(
        valid_inputs=[{DataType.STRING}, {DataType.STRING}, {DataType.INTEGER}],
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.STRING,
        arg_count=3,
    ),
    FunctionType.REGEXP_REPLACE: FunctionConfig(
        valid_inputs={DataType.STRING},
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.STRING,
        arg_count=3,
    ),
    FunctionType.DATE: FunctionConfig(
        valid_inputs={
            DataType.DATE,
            DataType.TIMESTAMP,
            DataType.DATETIME,
            DataType.STRING,
        },
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.DATE,
        arg_count=1,
    ),
    FunctionType.DATE_TRUNCATE: FunctionConfig(
        valid_inputs=[
            {
                DataType.DATE,
                DataType.TIMESTAMP,
                DataType.DATETIME,
                DataType.STRING,
            },
            {DataType.DATE_PART},
        ],
        output_purpose=Purpose.PROPERTY,
        # output_type=DataType.DATE,
        output_type_function=get_date_trunc_output,
        arg_count=2,
    ),
    FunctionType.DATE_PART: FunctionConfig(
        valid_inputs=[
            {
                DataType.DATE,
                DataType.TIMESTAMP,
                DataType.DATETIME,
                DataType.STRING,
            },
            {DataType.DATE_PART},
        ],
        output_purpose=Purpose.PROPERTY,
        output_type_function=get_date_part_output,
        arg_count=2,
    ),
    FunctionType.DATE_ADD: FunctionConfig(
        valid_inputs=[
            {
                DataType.DATE,
                DataType.TIMESTAMP,
                DataType.DATETIME,
                DataType.STRING,
            },
            {DataType.DATE_PART},
            {DataType.INTEGER},
        ],
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.DATE,
        arg_count=3,
    ),
    FunctionType.DATE_SUB: FunctionConfig(
        valid_inputs=[
            {
                DataType.DATE,
                DataType.TIMESTAMP,
                DataType.DATETIME,
                DataType.STRING,
            },
            {DataType.DATE_PART},
            {DataType.INTEGER},
        ],
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.DATE,
        arg_count=3,
    ),
    FunctionType.DATE_DIFF: FunctionConfig(
        valid_inputs=[
            {
                DataType.DATE,
                DataType.TIMESTAMP,
                DataType.DATETIME,
                DataType.STRING,
            },
            {
                DataType.DATE,
                DataType.TIMESTAMP,
                DataType.DATETIME,
                DataType.STRING,
            },
            {DataType.DATE_PART},
        ],
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.INTEGER,
        arg_count=3,
    ),
    FunctionType.DATETIME: FunctionConfig(
        valid_inputs={
            DataType.DATE,
            DataType.TIMESTAMP,
            DataType.DATETIME,
            DataType.STRING,
        },
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.DATETIME,
        arg_count=1,
    ),
    FunctionType.TIMESTAMP: FunctionConfig(
        valid_inputs={
            DataType.DATE,
            DataType.TIMESTAMP,
            DataType.DATETIME,
            DataType.STRING,
        },
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.TIMESTAMP,
        arg_count=1,
    ),
    FunctionType.SECOND: FunctionConfig(
        valid_inputs={
            DataType.DATE,
            DataType.TIMESTAMP,
            DataType.DATETIME,
            DataType.STRING,
        },
        output_purpose=Purpose.PROPERTY,
        output_type=TraitDataType(type=DataType.INTEGER, traits=["second"]),
        arg_count=1,
    ),
    FunctionType.MINUTE: FunctionConfig(
        valid_inputs={
            DataType.DATE,
            DataType.TIMESTAMP,
            DataType.DATETIME,
            DataType.STRING,
        },
        output_purpose=Purpose.PROPERTY,
        output_type=TraitDataType(type=DataType.INTEGER, traits=["minute"]),
        arg_count=1,
    ),
    FunctionType.HOUR: FunctionConfig(
        valid_inputs={
            DataType.DATE,
            DataType.TIMESTAMP,
            DataType.DATETIME,
            DataType.STRING,
        },
        output_purpose=Purpose.PROPERTY,
        output_type=TraitDataType(type=DataType.INTEGER, traits=["hour"]),
        arg_count=1,
    ),
    FunctionType.DAY: FunctionConfig(
        valid_inputs={
            DataType.DATE,
            DataType.TIMESTAMP,
            DataType.DATETIME,
            DataType.STRING,
        },
        output_purpose=Purpose.PROPERTY,
        output_type=TraitDataType(type=DataType.INTEGER, traits=["day"]),
        arg_count=1,
    ),
    FunctionType.WEEK: FunctionConfig(
        valid_inputs={
            DataType.DATE,
            DataType.TIMESTAMP,
            DataType.DATETIME,
            DataType.STRING,
        },
        output_purpose=Purpose.PROPERTY,
        output_type=TraitDataType(type=DataType.INTEGER, traits=["week"]),
        arg_count=1,
    ),
    FunctionType.MONTH: FunctionConfig(
        valid_inputs={
            DataType.DATE,
            DataType.TIMESTAMP,
            DataType.DATETIME,
            DataType.STRING,
        },
        output_purpose=Purpose.PROPERTY,
        output_type=TraitDataType(type=DataType.INTEGER, traits=["month"]),
        arg_count=1,
    ),
    FunctionType.QUARTER: FunctionConfig(
        valid_inputs={
            DataType.DATE,
            DataType.TIMESTAMP,
            DataType.DATETIME,
            DataType.STRING,
        },
        output_purpose=Purpose.PROPERTY,
        output_type=TraitDataType(type=DataType.INTEGER, traits=["quarter"]),
        arg_count=1,
    ),
    FunctionType.YEAR: FunctionConfig(
        valid_inputs={
            DataType.DATE,
            DataType.TIMESTAMP,
            DataType.DATETIME,
            DataType.STRING,
        },
        output_purpose=Purpose.PROPERTY,
        output_type=TraitDataType(type=DataType.INTEGER, traits=["year"]),
        arg_count=1,
    ),
    FunctionType.DAY_OF_WEEK: FunctionConfig(
        valid_inputs={
            DataType.DATE,
            DataType.TIMESTAMP,
            DataType.DATETIME,
            DataType.STRING,
        },
        output_purpose=Purpose.PROPERTY,
        output_type=TraitDataType(type=DataType.INTEGER, traits=["day_of_week"]),
        arg_count=1,
    ),
    FunctionType.ADD: FunctionConfig(
        valid_inputs={
            DataType.INTEGER,
            DataType.FLOAT,
            DataType.NUMBER,
            DataType.NUMERIC,
        },
        output_purpose=Purpose.PROPERTY,
        arg_count=InfiniteFunctionArgs,
    ),
    FunctionType.SUBTRACT: FunctionConfig(
        valid_inputs={
            DataType.INTEGER,
            DataType.FLOAT,
            DataType.NUMBER,
            DataType.NUMERIC,
        },
        output_purpose=Purpose.PROPERTY,
        arg_count=InfiniteFunctionArgs,
    ),
    FunctionType.MULTIPLY: FunctionConfig(
        valid_inputs={
            DataType.INTEGER,
            DataType.FLOAT,
            DataType.NUMBER,
            DataType.NUMERIC,
        },
        output_purpose=Purpose.PROPERTY,
        arg_count=InfiniteFunctionArgs,
    ),
    FunctionType.DIVIDE: FunctionConfig(
        valid_inputs={
            DataType.INTEGER,
            DataType.FLOAT,
            DataType.NUMBER,
            DataType.NUMERIC,
        },
        output_purpose=Purpose.PROPERTY,
        arg_count=InfiniteFunctionArgs,
    ),
    FunctionType.MOD: FunctionConfig(
        valid_inputs=[
            {DataType.INTEGER, DataType.FLOAT, DataType.NUMBER, DataType.NUMERIC},
            {DataType.INTEGER},
        ],
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.INTEGER,
        arg_count=2,
    ),
    FunctionType.SQRT: FunctionConfig(
        valid_inputs=[
            {DataType.INTEGER, DataType.FLOAT, DataType.NUMBER, DataType.NUMERIC},
        ],
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.INTEGER,
        arg_count=1,
    ),
    FunctionType.LOG: FunctionConfig(
        valid_inputs=[
            {DataType.INTEGER, DataType.FLOAT, DataType.NUMBER, DataType.NUMERIC},
            {DataType.INTEGER},
        ],
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.FLOAT,
        arg_count=2,
    ),
    FunctionType.RANDOM: FunctionConfig(
        valid_inputs=[],
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.FLOAT,
        arg_count=1,
    ),
    FunctionType.ROUND: FunctionConfig(
        valid_inputs=[
            {DataType.INTEGER, DataType.FLOAT, DataType.NUMBER, DataType.NUMERIC},
            {DataType.INTEGER},
        ],
        output_purpose=Purpose.PROPERTY,
        output_type_function=lambda args: get_output_type_at_index(args, 0),
        arg_count=2,
    ),
    FunctionType.FLOOR: FunctionConfig(
        valid_inputs=[
            {DataType.INTEGER, DataType.FLOAT, DataType.NUMBER, DataType.NUMERIC},
        ],
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.INTEGER,
        arg_count=1,
    ),
    FunctionType.CEIL: FunctionConfig(
        valid_inputs=[
            {DataType.INTEGER, DataType.FLOAT, DataType.NUMBER, DataType.NUMERIC},
        ],
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.INTEGER,
        arg_count=1,
    ),
    FunctionType.CUSTOM: FunctionConfig(
        output_purpose=Purpose.PROPERTY,
        arg_count=InfiniteFunctionArgs,
    ),
    FunctionType.CASE: FunctionConfig(
        output_purpose=Purpose.PROPERTY,
        output_type_function=validate_case_output,
        arg_count=InfiniteFunctionArgs,
    ),
    FunctionType.CAST: FunctionConfig(
        output_purpose=Purpose.PROPERTY,
        arg_count=2,
        output_type_function=get_cast_output_type,
    ),
    FunctionType.CONCAT: FunctionConfig(
        valid_inputs={DataType.STRING},
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.STRING,
        arg_count=InfiniteFunctionArgs,
    ),
    FunctionType.CONSTANT: FunctionConfig(
        output_purpose=Purpose.CONSTANT,
        arg_count=1,
    ),
    FunctionType.TYPED_CONSTANT: FunctionConfig(
        output_purpose=Purpose.CONSTANT,
        output_type_function=get_cast_output_type,
        arg_count=2,
    ),
    FunctionType.IS_NULL: FunctionConfig(
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.BOOL,
        arg_count=1,
    ),
    FunctionType.STRUCT: FunctionConfig(
        output_purpose=Purpose.PROPERTY,
        arg_count=InfiniteFunctionArgs,
        output_type_function=create_struct_output,
    ),
    FunctionType.ARRAY: FunctionConfig(
        output_purpose=Purpose.PROPERTY,
        arg_count=InfiniteFunctionArgs,
        output_type=ArrayType(type=DataType.STRING),
    ),
    FunctionType.LENGTH: FunctionConfig(
        valid_inputs={DataType.STRING, DataType.ARRAY, DataType.MAP},
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.INTEGER,
        arg_count=1,
    ),
    FunctionType.SUM: FunctionConfig(
        valid_inputs={
            DataType.INTEGER,
            DataType.FLOAT,
            DataType.NUMBER,
            DataType.NUMERIC,
            DataType.BOOL,
        },
        output_purpose=Purpose.METRIC,
        arg_count=1,
    ),
    FunctionType.ARRAY_AGG: FunctionConfig(
        valid_inputs={*DataType},
        output_purpose=Purpose.METRIC,
        output_type_function=lambda args: ArrayType(
            type=merge_datatypes([arg_to_datatype(x) for x in args])
        ),
        arg_count=1,
    ),
    FunctionType.AVG: FunctionConfig(
        valid_inputs={
            DataType.INTEGER,
            DataType.FLOAT,
            DataType.NUMBER,
            DataType.NUMERIC,
        },
        output_purpose=Purpose.METRIC,
        arg_count=1,
    ),
    FunctionType.UNIX_TO_TIMESTAMP: FunctionConfig(
        valid_inputs={DataType.INTEGER},
        output_purpose=Purpose.PROPERTY,
        output_type=DataType.TIMESTAMP,
        arg_count=1,
    ),
}

EXCLUDED_FUNCTIONS = {
    FunctionType.CUSTOM,
    # Temporary
    FunctionType.DATE_LITERAL,
    FunctionType.DATETIME_LITERAL,
    FunctionType.ARRAY,
}

for k in FunctionType.__members__.values():
    if k not in FUNCTION_REGISTRY and k not in EXCLUDED_FUNCTIONS:
        raise InvalidSyntaxException(
            f"Function enum value {k} not in creation registry"
        )


class FunctionFactory:
    def __init__(self, environment: Environment | None = None):
        self.environment = environment

    def create_function(
        self,
        args: list[Any],
        operator: FunctionType,
        meta: Meta | None = None,
    ):
        if operator not in FUNCTION_REGISTRY:
            raise ValueError(f"Function {operator} not in registry")
        config = FUNCTION_REGISTRY[operator]
        valid_inputs: set[DataType] | list[set[DataType]] = config.valid_inputs or set(
            DataType
        )
        output_purpose = config.output_purpose
        base_output_type = config.output_type
        arg_count = config.arg_count

        if args:
            if not self.environment:
                raise ValueError("Environment required for function creation with args")
            # TODO: remove this dependency
            from trilogy.parsing.common import process_function_args

            full_args = process_function_args(
                args, environment=self.environment, meta=meta
            )
        else:
            full_args = []
        final_output_type: CONCRETE_TYPES
        if config.output_type_function:

            final_output_type = config.output_type_function(full_args)
        elif not base_output_type:

            final_output_type = merge_datatypes([arg_to_datatype(x) for x in full_args])
        elif base_output_type:
            final_output_type = base_output_type
        else:
            raise SyntaxError(f"Could not determine output type for {operator}")
        if not output_purpose:
            if operator in FunctionClass.AGGREGATE_FUNCTIONS.value:
                output_purpose = Purpose.METRIC
            else:
                output_purpose = Purpose.PROPERTY

        return Function(
            operator=operator,
            arguments=full_args,  # type: ignore
            output_datatype=final_output_type,
            output_purpose=output_purpose,
            valid_inputs=valid_inputs,
            arg_count=arg_count,
        )


def create_function_derived_concept(
    name: str,
    namespace: str,
    operator: FunctionType,
    arguments: list[Concept],
    environment: Environment,
    output_type: Optional[
        DataType | ArrayType | StructType | MapType | NumericType | TraitDataType
    ] = None,
    output_purpose: Optional[Purpose] = None,
) -> Concept:
    purpose = (
        function_args_to_output_purpose(arguments, environment=environment)
        if output_purpose is None
        else output_purpose
    )
    output_type = arg_to_datatype(arguments[0]) if output_type is None else output_type
    return Concept(
        name=name,
        namespace=namespace,
        datatype=output_type,
        purpose=purpose,
        lineage=Function(
            operator=operator,
            arguments=[x.reference for x in arguments],
            output_datatype=output_type,
            output_purpose=purpose,
            arg_count=len(arguments),
        ),
    )


def argument_to_purpose(arg) -> Purpose:
    if isinstance(arg, UndefinedConcept):
        return Purpose.UNKNOWN
    if isinstance(arg, Function):
        return arg.output_purpose
    elif isinstance(arg, AggregateWrapper):
        base = arg.function.output_purpose
        if arg.by and base == Purpose.METRIC:
            return Purpose.PROPERTY
        return arg.function.output_purpose
    elif isinstance(arg, Parenthetical):
        return argument_to_purpose(arg.content)
    elif isinstance(arg, WindowItem):
        return Purpose.PROPERTY
    elif isinstance(arg, Conditional):
        return Purpose.PROPERTY
    elif isinstance(arg, Concept):
        base = arg.purpose
        if (
            isinstance(arg.lineage, AggregateWrapper)
            and arg.lineage.by
            and base == Purpose.METRIC
        ):
            return Purpose.PROPERTY
        return arg.purpose
    elif isinstance(arg, (int, float, str, bool, list, NumericType, DataType)):
        return Purpose.CONSTANT
    elif isinstance(arg, DatePart):
        return Purpose.CONSTANT
    elif isinstance(arg, MagicConstants):
        return Purpose.CONSTANT
    else:
        raise ValueError(f"Cannot parse arg purpose for {arg} of type {type(arg)}")


def function_args_to_output_purpose(args, environment: Environment) -> Purpose:
    has_metric = False
    has_non_constant = False
    has_non_single_row_constant = False
    if not args:
        return Purpose.CONSTANT
    for arg in args:
        if isinstance(arg, ConceptRef):
            arg = environment.concepts[arg.address]
        purpose = argument_to_purpose(arg)
        if purpose == Purpose.METRIC:
            has_metric = True
        if purpose != Purpose.CONSTANT:
            has_non_constant = True
        if isinstance(arg, Concept) and arg.granularity != Granularity.SINGLE_ROW:
            has_non_single_row_constant = True
    if args and not has_non_constant and not has_non_single_row_constant:
        return Purpose.CONSTANT
    if has_metric:
        return Purpose.METRIC
    return Purpose.PROPERTY


def Count(args: list[Concept], environment: Environment) -> Function:
    return FunctionFactory(environment).create_function(
        args=args, operator=FunctionType.COUNT
    )


def CountDistinct(args: list[Concept], environment: Environment) -> Function:
    return FunctionFactory(environment).create_function(
        args=args, operator=FunctionType.COUNT
    )


def Max(args: list[Concept], environment: Environment) -> Function:
    return FunctionFactory(environment).create_function(
        args=args, operator=FunctionType.COUNT
    )


def Min(args: list[Concept], environment: Environment) -> Function:
    return FunctionFactory(environment).create_function(
        args=args, operator=FunctionType.COUNT
    )


def Split(args: list[Concept], environment: Environment) -> Function:
    return FunctionFactory(environment).create_function(
        args=args, operator=FunctionType.SPLIT
    )


def AttrAccess(args: list[ConceptRef | str | int], environment: Environment):
    return FunctionFactory(environment).create_function(
        args=args, operator=FunctionType.ATTR_ACCESS
    )


def CurrentDate(
    args: list[Concept], environment: Environment | None = None
) -> Function:
    return FunctionFactory(environment).create_function(
        args=args, operator=FunctionType.CURRENT_DATE
    )


def CurrentDatetime(
    args: list[Concept], environment: Environment | None = None
) -> Function:
    return FunctionFactory(environment).create_function(
        args=args, operator=FunctionType.CURRENT_DATETIME
    )



from trilogy import Dialects
from trilogy.core.statements.author import ShowStatement
from trilogy.parser import parse


def test_show_bigquery():
    declarations = """
    key user_id int metadata(description="the description");
    property user_id.display_name string metadata(description="The display name ");
    property user_id.about_me string metadata(description="User provided description");
    key post_id int;


    datasource posts (
        user_id: user_id,
        id: post_id
        )
        grain (post_id)
        address `bigquery-public-data.stackoverflow.post_history`
    ;


    datasource users (
        id: user_id,
        display_name: display_name,
        about_me: about_me,
        )
        grain (user_id)
        address `bigquery-public-data.stackoverflow.users`
    ;


        """
    env, parsed = parse(declarations)

    q1 = """
    metric post_count<- count(post_id);
    metric distinct_post_count <- count_distinct(post_id);
    
    metric user_count <- count(user_id);
    
    show select
        post_count,
        distinct_post_count,
        user_count
    ;"""
    env, parsed = parse(q1, environment=env)
    select: ShowStatement = parsed[-1]

    query = (
        Dialects.DUCK_DB.default_executor(environment=env)
        .execute_query(select)
        .fetchall()
    )
    assert (
        'FULL JOIN "cheerful" on 1=1' in query[0]["__preql_internal_query_text"]
    ), query[0]["__preql_internal_query_text"]



from dataclasses import dataclass
from typing import Any, List, Optional

from trilogy.core.models.author import ConceptRef
from trilogy.core.models.datasource import Datasource
from trilogy.core.models.environment import Environment
from trilogy.core.statements.author import (
    ConceptDeclarationStatement,
    ImportStatement,
    MergeStatementV2,
)
from trilogy.core.statements.execute import (
    ProcessedShowStatement,
    ProcessedStaticValueOutput,
    ProcessedValidateStatement,
)
from trilogy.core.validation.common import ValidationTest
from trilogy.dialect.base import BaseDialect
from trilogy.engine import ResultProtocol


@dataclass
class MockResult(ResultProtocol):
    values: list["MockResultRow"]
    columns: list[str]

    def __init__(self, values: list[Any], columns: list[str]):
        processed: list[MockResultRow] = []
        for x in values:
            if isinstance(x, dict):
                processed.append(MockResultRow(x))
            elif isinstance(x, MockResultRow):
                processed.append(x)
            else:
                raise ValueError(
                    f"Cannot process value of type {type(x)} in MockResult"
                )
        self.columns = columns
        self.values = processed

    def __iter__(self):
        while self.values:
            yield self.values.pop(0)

    def fetchall(self):
        return self.values

    def fetchone(self):
        if self.values:
            return self.values.pop(0)
        return None

    def fetchmany(self, size: int):
        rval = self.values[:size]
        self.values = self.values[size:]
        return rval

    def keys(self):
        return self.columns

    def as_dict(self):
        return [x.as_dict() if isinstance(x, MockResultRow) else x for x in self.values]


@dataclass
class MockResultRow:
    _values: dict[str, Any]

    def as_dict(self):
        return self._values

    def __str__(self) -> str:
        return str(self._values)

    def __repr__(self) -> str:
        return repr(self._values)

    def __getattr__(self, name: str) -> Any:
        if name in self._values:
            return self._values[name]
        return super().__getattribute__(name)

    def __getitem__(self, key: str) -> Any:
        return self._values[key]

    def values(self):
        return self._values.values()

    def keys(self):
        return self._values.keys()


def generate_result_set(
    columns: List[ConceptRef], output_data: list[Any]
) -> MockResult:
    """Generate a mock result set from columns and output data."""
    names = [x.address.replace(".", "_") for x in columns]
    return MockResult(
        values=[dict(zip(names, [row])) for row in output_data], columns=names
    )


def handle_concept_declaration(query: ConceptDeclarationStatement) -> MockResult:
    """Handle concept declaration statements without execution."""
    concept = query.concept
    return MockResult(
        [
            {
                "address": concept.address,
                "type": concept.datatype.value,
                "purpose": concept.purpose.value,
                "derivation": concept.derivation.value,
            }
        ],
        ["address", "type", "purpose", "derivation"],
    )


def handle_datasource(query: Datasource) -> MockResult:
    """Handle datasource queries without execution."""
    return MockResult(
        [
            {
                "name": query.name,
            }
        ],
        ["name"],
    )


def handle_import_statement(query: ImportStatement) -> MockResult:
    """Handle import statements without execution."""
    return MockResult(
        [
            {
                "path": query.path,
                "alias": query.alias,
            }
        ],
        ["path", "alias"],
    )


def handle_merge_statement(
    query: MergeStatementV2, environment: Environment
) -> MockResult:
    """Handle merge statements by updating environment and returning result."""
    for concept in query.sources:
        environment.merge_concept(
            concept, query.targets[concept.address], modifiers=query.modifiers
        )

    return MockResult(
        [
            {
                "sources": ",".join([x.address for x in query.sources]),
                "targets": ",".join([x.address for _, x in query.targets.items()]),
            }
        ],
        ["source", "target"],
    )


def handle_processed_show_statement(
    query: ProcessedShowStatement, compiled_statements: list[str]
) -> MockResult:
    """Handle processed show statements without execution."""

    return generate_result_set(query.output_columns, compiled_statements)


def raw_validation_to_result(
    raw: list[ValidationTest], generator: Optional[BaseDialect] = None
) -> Optional[MockResult]:
    """Convert raw validation tests to mock result."""
    if not raw:
        return MockResult([], ["check_type", "expected", "result", "ran", "query"])
    output = []
    for row in raw:
        if row.raw_query and generator and not row.generated_query:
            try:
                row.generated_query = generator.compile_statement(row.raw_query)
            except Exception as e:
                row.generated_query = f"Error generating query: {e}"
        output.append(
            {
                "check_type": row.check_type.value,
                "expected": row.expected,
                "result": str(row.result) if row.result else None,
                "ran": row.ran,
                "query": row.generated_query if row.generated_query else "",
            }
        )
    return MockResult(output, ["check_type", "expected", "result", "ran", "query"])


def handle_processed_validate_statement(
    query: ProcessedValidateStatement, dialect: BaseDialect, validate_environment_func
) -> Optional[MockResult]:
    """Handle processed validate statements."""
    results = validate_environment_func(query.scope, query.targets)
    return raw_validation_to_result(results, dialect)


def handle_show_statement_outputs(
    statement: ProcessedShowStatement,
    compiled_statements: list[str],
    environment: Environment,
    dialect: BaseDialect,
) -> list[MockResult]:
    """Handle show statement outputs without execution."""
    output = []
    for x in statement.output_values:
        if isinstance(x, ProcessedStaticValueOutput):
            output.append(generate_result_set(statement.output_columns, x.values))
        elif compiled_statements:

            output.append(
                generate_result_set(
                    statement.output_columns,
                    compiled_statements,
                )
            )
        elif isinstance(x, ProcessedValidateStatement):
            from trilogy.core.validation.environment import validate_environment

            raw = validate_environment(environment, x.scope, x.targets)
            results = raw_validation_to_result(raw, dialect)
            if results:
                output.append(results)
        else:
            raise NotImplementedError(f"Cannot show type {type(x)} in show statement")
    return output



from typing import List, Optional

from trilogy.constants import logger
from trilogy.core.constants import CONSTANT_DATASET
from trilogy.core.enums import Derivation, Purpose, SourceType
from trilogy.core.models.build import (
    BuildComparison,
    BuildConcept,
    BuildConditional,
    BuildDatasource,
    BuildFunction,
    BuildGrain,
    BuildOrderBy,
    BuildParenthetical,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.models.execute import QueryDatasource, UnnestJoin
from trilogy.core.processing.nodes.base_node import StrategyNode, resolve_concept_map
from trilogy.utility import unique

LOGGER_PREFIX = "[CONCEPT DETAIL - SELECT NODE]"


class SelectNode(StrategyNode):
    """Select nodes actually fetch raw data from a table
    Responsible for selecting the cheapest option from which to select.
    """

    source_type = SourceType.SELECT

    def __init__(
        self,
        input_concepts: List[BuildConcept],
        output_concepts: List[BuildConcept],
        environment: BuildEnvironment,
        datasource: BuildDatasource | None = None,
        whole_grain: bool = False,
        parents: List["StrategyNode"] | None = None,
        depth: int = 0,
        partial_concepts: List[BuildConcept] | None = None,
        nullable_concepts: List[BuildConcept] | None = None,
        accept_partial: bool = False,
        grain: Optional[BuildGrain] = None,
        force_group: bool | None = False,
        conditions: (
            BuildConditional | BuildComparison | BuildParenthetical | None
        ) = None,
        preexisting_conditions: (
            BuildConditional | BuildComparison | BuildParenthetical | None
        ) = None,
        hidden_concepts: set[str] | None = None,
        ordering: BuildOrderBy | None = None,
    ):
        super().__init__(
            input_concepts=input_concepts,
            output_concepts=output_concepts,
            environment=environment,
            whole_grain=whole_grain,
            parents=parents,
            depth=depth,
            partial_concepts=partial_concepts,
            nullable_concepts=nullable_concepts,
            force_group=force_group,
            grain=grain,
            conditions=conditions,
            preexisting_conditions=preexisting_conditions,
            hidden_concepts=hidden_concepts,
            ordering=ordering,
        )
        self.accept_partial = accept_partial
        self.datasource = datasource

    def validate_inputs(self):
        # we do not need to validate inputs for a select node
        # as it will be a root
        return

    def resolve_from_provided_datasource(
        self,
    ) -> QueryDatasource:
        if not self.datasource:
            raise ValueError("Datasource not provided")
        datasource: BuildDatasource = self.datasource

        all_concepts_final: List[BuildConcept] = unique(self.all_concepts, "address")
        source_map: dict[str, set[BuildDatasource | QueryDatasource | UnnestJoin]] = {
            concept.address: {datasource} for concept in self.input_concepts
        }

        derived_concepts = [
            c
            for c in datasource.columns
            if isinstance(c.alias, BuildFunction) and c.concept.address in source_map
        ]
        for c in derived_concepts:
            if not isinstance(c.alias, BuildFunction):
                continue
            for x in c.alias.concept_arguments:
                source_map[x.address] = {datasource}
        for x in all_concepts_final:
            if x.address not in source_map and x.derivation in (
                Derivation.MULTISELECT,
                Derivation.FILTER,
                Derivation.BASIC,
                Derivation.ROWSET,
                Derivation.BASIC,
                Derivation.UNION,
            ):
                source_map[x.address] = set()

        # if we're not grouping
        # force grain to datasource grain
        # so that we merge on the same grain
        if self.force_group is False:
            grain = datasource.grain
        else:
            grain = self.grain or BuildGrain()
        return QueryDatasource(
            input_concepts=self.input_concepts,
            output_concepts=all_concepts_final,
            source_map=source_map,
            datasources=[datasource],
            grain=grain,
            joins=[],
            partial_concepts=[
                c.concept for c in datasource.columns if not c.is_complete
            ],
            nullable_concepts=[c.concept for c in datasource.columns if c.is_nullable],
            source_type=SourceType.DIRECT_SELECT,
            # we can skip rendering conditions
            condition=self.conditions,
            # select nodes should never group
            force_group=self.force_group,
            hidden_concepts=self.hidden_concepts,
            ordering=self.ordering,
        )

    def resolve_from_constant_datasources(self) -> QueryDatasource:
        datasource = BuildDatasource(
            name=CONSTANT_DATASET, address=CONSTANT_DATASET, columns=[]
        )
        return QueryDatasource(
            input_concepts=[],
            output_concepts=unique(self.all_concepts, "address"),
            source_map={concept.address: set() for concept in self.all_concepts},
            datasources=[datasource],
            grain=datasource.grain,
            condition=self.conditions,
            joins=[],
            partial_concepts=[],
            source_type=SourceType.CONSTANT,
            hidden_concepts=self.hidden_concepts,
            ordering=self.ordering,
        )

    def _resolve(self) -> QueryDatasource:
        # if we have parent nodes, we do not need to go to a datasource
        resolution: QueryDatasource | None = None
        if all(
            [
                (
                    c.derivation == Derivation.CONSTANT
                    or (
                        c.purpose == Purpose.CONSTANT
                        and c.derivation == Derivation.MULTISELECT
                    )
                )
                for c in self.all_concepts
            ]
        ):
            logger.info(
                f"{self.logging_prefix}{LOGGER_PREFIX} have a constant datasource"
            )
            resolution = self.resolve_from_constant_datasources()
            return resolution

        if self.datasource and not resolution:
            resolution = self.resolve_from_provided_datasource()

        if self.parents:
            if not resolution:
                return super()._resolve()
            # zip in our parent source map
            parent_sources: List[QueryDatasource | BuildDatasource] = [
                p.resolve() for p in self.parents
            ]

            resolution.datasources += parent_sources

            source_map = resolve_concept_map(
                parent_sources,
                targets=self.output_concepts,
                inherited_inputs=self.input_concepts + self.existence_concepts,
            )
            for k, v in source_map.items():
                if v and k not in resolution.source_map:
                    resolution.source_map[k] = v
        if not resolution:
            raise ValueError(f"No select node could be generated for {self}")
        return resolution

    def copy(self) -> "SelectNode":
        return SelectNode(
            input_concepts=list(self.input_concepts),
            output_concepts=list(self.output_concepts),
            environment=self.environment,
            datasource=self.datasource,
            depth=self.depth,
            parents=self.parents,
            whole_grain=self.whole_grain,
            partial_concepts=list(self.partial_concepts),
            nullable_concepts=list(self.nullable_concepts),
            accept_partial=self.accept_partial,
            grain=self.grain,
            force_group=self.force_group,
            conditions=self.conditions,
            preexisting_conditions=self.preexisting_conditions,
            hidden_concepts=self.hidden_concepts,
            ordering=self.ordering,
        )


class ConstantNode(SelectNode):
    source_type = SourceType.CONSTANT
    """Represents a constant value."""

    def copy(self) -> "ConstantNode":
        return ConstantNode(
            input_concepts=list(self.input_concepts),
            output_concepts=list(self.output_concepts),
            environment=self.environment,
            datasource=self.datasource,
            depth=self.depth,
            partial_concepts=list(self.partial_concepts),
            conditions=self.conditions,
            preexisting_conditions=self.preexisting_conditions,
            hidden_concepts=self.hidden_concepts,
            ordering=self.ordering,
        )

    def _resolve(self) -> QueryDatasource:
        return self.resolve_from_constant_datasources()



from pytest import raises

from trilogy.core.exceptions import InvalidSyntaxException
from trilogy.core.models.execute import (
    BuildConcept,
    BuildDatasource,
    BuildGrain,
    DataType,
    Purpose,
    raise_helpful_join_validation_error,
)


def test_raise_helpful_join_validation_error():

    with raises(InvalidSyntaxException):
        raise_helpful_join_validation_error(
            concepts=[
                BuildConcept(
                    name="test_concept",
                    datatype=DataType.INTEGER,
                    purpose=Purpose.KEY,
                    build_is_aggregate=False,
                    grain=BuildGrain(),
                )
            ],
            left_datasource=BuildDatasource(name="left_ds", columns=[], address="agsg"),
            right_datasource=BuildDatasource(
                name="right_ds", columns=[], address="agsg"
            ),
        )
    with raises(InvalidSyntaxException):
        raise_helpful_join_validation_error(
            concepts=[
                BuildConcept(
                    name="test_concept",
                    datatype=DataType.INTEGER,
                    purpose=Purpose.KEY,
                    build_is_aggregate=False,
                    grain=BuildGrain(),
                )
            ],
            left_datasource=None,
            right_datasource=BuildDatasource(
                name="right_ds", columns=[], address="agsg"
            ),
        )
    with raises(InvalidSyntaxException):
        raise_helpful_join_validation_error(
            concepts=[
                BuildConcept(
                    name="test_concept",
                    datatype=DataType.INTEGER,
                    purpose=Purpose.KEY,
                    build_is_aggregate=False,
                    grain=BuildGrain(),
                )
            ],
            left_datasource=BuildDatasource(name="left_ds", columns=[], address="agsg"),
            right_datasource=None,
        )



CONSTANT_DATASET: str = "preql_internal_constant_dataset"
ALL_ROWS_CONCEPT = "all_rows"
INTERNAL_NAMESPACE = "__preql_internal"
PERSISTED_CONCEPT_PREFIX = "__pre_persist"
UNNEST_NAME = "_unnest_alias"



from typing import List

from trilogy.constants import logger
from trilogy.core.models.build import BuildConcept, BuildFunction, BuildWhereClause
from trilogy.core.models.build_environment import BuildEnvironment

# C:\Users\ethan\coding_projects\pytrilogy\trilogy\core\processing\node_generators\group_to_node.py
from trilogy.core.processing.nodes import (
    GroupNode,
    History,
    MergeNode,
    StrategyNode,
)
from trilogy.core.processing.utility import padding

LOGGER_PREFIX = "[GEN_GROUP_TO_NODE]"


def gen_group_to_node(
    concept: BuildConcept,
    local_optional,
    environment: BuildEnvironment,
    g,
    depth: int,
    source_concepts,
    history: History | None = None,
    conditions: BuildWhereClause | None = None,
) -> GroupNode | MergeNode:
    # aggregates MUST always group to the proper grain
    if not isinstance(concept.lineage, BuildFunction):
        raise SyntaxError(
            f"Group to should have function lineage, is {type(concept.lineage)}"
        )

    parent_concepts: List[BuildConcept] = concept.lineage.concept_arguments
    root = parent_concepts[0]
    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} group by node has required parents {[x.address for x in parent_concepts]}"
    )
    parents: List[StrategyNode] = [
        source_concepts(
            mandatory_list=parent_concepts,
            environment=environment,
            g=g,
            depth=depth + 1,
            history=history,
            conditions=conditions,
        )
    ]
    outputs = parent_concepts + [concept]
    group_node = GroupNode(
        output_concepts=parent_concepts + [concept],
        input_concepts=parent_concepts,
        environment=environment,
        parents=parents,
        depth=depth,
        preexisting_conditions=conditions.conditional if conditions else None,
        hidden_concepts=set(
            [
                x.address
                for x in outputs
                if x.address not in local_optional
                and x.address != concept.address
                and x.address != root.address
            ]
        ),
    )

    # early exit if no optional
    missing_local_option: list[BuildConcept] = [
        x for x in local_optional if x not in group_node.output_concepts
    ]
    if not missing_local_option:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} no missing local optional required, returning group node only."
        )
        return group_node

    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} group by node is missing required optional {[x.address for x in missing_local_option]}"
    )

    # the keys we group by
    # are what we can use for enrichment
    enrich_node = source_concepts(  # this fetches the parent + join keys
        # to then connect to the rest of the query
        mandatory_list=parent_concepts + local_optional,
        environment=environment,
        g=g,
        depth=depth + 1,
        history=history,
        conditions=conditions,
    )
    if not enrich_node:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} group by node enrich node, returning group node only."
        )
        return group_node
    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} returning group to node with enrichment."
    )
    return MergeNode(
        input_concepts=[concept]
        + local_optional
        + [x for x in parent_concepts if x.address != concept.address],
        output_concepts=[concept] + local_optional,
        environment=environment,
        parents=[
            # this node gets the group
            group_node,
            # this node gets enrichment
            enrich_node,
        ],
        whole_grain=True,
        depth=depth,
        preexisting_conditions=conditions.conditional if conditions else None,
    )



from trilogy.core.enums import Purpose
from trilogy.core.exceptions import UndefinedConceptException
from trilogy.core.models.author import Concept
from trilogy.core.models.core import DataType
from trilogy.core.models.environment import EnvironmentConceptDict
from trilogy.parser import parse


def test_undefined_concept_query(test_environment):
    q = "SELECT orid LIMIT 10;"
    try:
        parse(q, test_environment)
    except UndefinedConceptException as e:
        assert e.suggestions == ["order_id"]

    q = "SELECT order_ct LIMIT 10;"
    try:
        parse(q, test_environment)
    except UndefinedConceptException as e:
        assert len(e.suggestions) == 3


def test_undefined_concept_dict():
    env = EnvironmentConceptDict()
    env["order_id"] = Concept(
        name="order_id", datatype=DataType.INTEGER, purpose=Purpose.KEY
    )
    try:
        env["zzz"]
    except UndefinedConceptException as e:
        assert e.suggestions == []
        assert "suggestions" not in e.message.lower()

    try:
        env["orid"]
    except UndefinedConceptException as e:
        assert e.suggestions == ["order_id"]
        assert "suggestions" in e.message.lower()
        assert "order_id" in e.message.lower()
